from __future__ import annotations

from collections.abc import Callable
from math import isclose
from typing import Union

from qadence.blocks import AbstractBlock, KronBlock, chain, kron, tag
from qadence.logger import get_logger
from qadence.operations import PHASE, RX, RY, RZ, H
from qadence.parameters import FeatureParameter, Parameter, VariationalParameter
from qadence.types import PI, BasisSet, ReuploadScaling, TParameter
from sympy import Basic, acos
from typing import Optional


logger = get_logger(__name__)

ROTATIONS = [RX, RY, RZ, PHASE]
RotationTypes = type[Union[RX, RY, RZ, PHASE]]


def _set_range(fm_type: BasisSet | Callable | str) -> tuple[float, float]:
    if fm_type == BasisSet.FOURIER:
        return (0.0, 2 * PI)
    elif fm_type == BasisSet.CHEBYSHEV:
        return (-1.0, 1.0)
    else:
        return (0.0, 1.0)


RS_FUNC_DICT = {
    ReuploadScaling.CONSTANT: lambda i: 1,
    ReuploadScaling.TOWER: lambda i: float(i + 1),
    ReuploadScaling.EXP: lambda i: float(2**i),
}


def fm_parameter_scaling(
    fm_type: BasisSet | Callable | str,
    param: Parameter | str = "phi",
    feature_range: tuple[float, float] | None = None,
    target_range: tuple[float, float] | None = None,
) -> Parameter | Basic:
    if isinstance(param, Parameter):
        fparam = param
        fparam.trainable = False
    else:
        fparam = FeatureParameter(param)

    # Set feature and target range
    feature_range = _set_range(fm_type) if feature_range is None else feature_range
    target_range = _set_range(fm_type) if target_range is None else target_range

    # Rescale the feature parameter
    scaling = (max(target_range) - min(target_range)) / (max(feature_range) - min(feature_range))
    shift = min(target_range) - min(feature_range) * scaling

    if isclose(scaling, 1.0):
        # So we don't get 1.0 factor in visualization
        scaled_fparam = fparam + shift
    else:
        scaled_fparam = scaling * fparam + shift

    return scaled_fparam


def fm_parameter_func(fm_type: BasisSet | Callable | str) -> Callable:
    def ident_fn(x: TParameter) -> TParameter:
        return x

    # Transform feature parameter
    if fm_type == BasisSet.FOURIER:
        transform_func = ident_fn
    elif fm_type == BasisSet.CHEBYSHEV:
        transform_func = acos
    elif callable(fm_type):
        transform_func = fm_type
    else:
        raise NotImplementedError(
            f"Feature map type {fm_type} not implemented. Choose an item from the BasisSet "
            f"enum: {BasisSet.list()}, or a custom defined sympy function to wrap "
            "the given feature parameter with."
        )

    return transform_func


def fm_reupload_scaling_fn(
    reupload_scaling: ReuploadScaling | Callable | str = ReuploadScaling.CONSTANT,
) -> tuple[Callable, str]:
    # Set reupload scaling function
    if callable(reupload_scaling):
        rs_func = reupload_scaling
        rs_tag = "Custom"
    else:
        rs_func = RS_FUNC_DICT.get(reupload_scaling, None)  # type: ignore [call-overload]
        if rs_func is None:
            raise NotImplementedError(
                f"Reupload scaling {reupload_scaling} not implemented; choose an item from "
                f"the ReuploadScaling enum: {[rs.name for rs in ReuploadScaling]}, or your own "
                "python function with a single int arg as input and int or float output."
            )
        if isinstance(reupload_scaling, ReuploadScaling):
            rs_tag = reupload_scaling.value
        else:
            rs_tag = reupload_scaling

    return rs_func, rs_tag


def feature_map(
    n_qubits: int,
    support: tuple[int, ...] | None = None,
    param: Parameter | str = "phi",
    op: RotationTypes = RX,
    fm_type: BasisSet | Callable | str = BasisSet.FOURIER,
    reupload_scaling: ReuploadScaling | Callable | str = ReuploadScaling.CONSTANT,
    feature_range: tuple[float, float] | None = None,
    target_range: tuple[float, float] | None = None,
    multiplier: Parameter | TParameter | None = None,
    param_prefix: str | None = None,
) -> KronBlock:
    """Construct a feature map of a given type.

    Arguments:
        n_qubits: Number of qubits the feature map covers. Results in `support=range(n_qubits)`.
        support: Puts one feature-encoding rotation gate on every qubit in `support`. n_qubits in
            this case specifies the total overall qubits of the circuit, which may be wider than the
            support itself, but not narrower.
        param: Parameter of the feature map; you can pass a string or Parameter;
            it will be set as non-trainable (FeatureParameter) regardless.
        op: Rotation operation of the feature map; choose from RX, RY, RZ or PHASE.
        fm_type: Basis set for data encoding; choose from `BasisSet.FOURIER` for Fourier
            encoding, or `BasisSet.CHEBYSHEV` for Chebyshev polynomials of the first kind.
        reupload_scaling: how the feature map scales the data that is re-uploaded for each qubit.
            choose from `ReuploadScaling` enumeration or provide your own function with a single
            int as input and int or float as output.
        feature_range: range of data that the input data provided comes from. Used to map input data
            to the correct domain of the feature-encoding function.
        target_range: range of data the data encoder assumes as the natural range. For example,
            in Chebyshev polynomials it is (-1, 1), while for Fourier it may be chosen as (0, 2*PI).
            Used to map data to the correct domain of the feature-encoding function.
        multiplier: overall multiplier; this is useful for reuploading the feature map serially with
            different scalings; can be a number or parameter/expression.
        param_prefix: string prefix to create trainable parameters multiplying the feature parameter
            inside the feature-encoding function. Note that currently this does not take into
            account the domain of the feature-encoding function.

    Example:
    ```python exec="on" source="material-block" result="json"
    from qadence_gradient import feature_map, BasisSet, ReuploadScaling

    fm = feature_map(3, fm_type=BasisSet.FOURIER)
    print(f"{fm = }")

    fm = feature_map(3, fm_type=BasisSet.CHEBYSHEV)
    print(f"{fm = }")

    fm = feature_map(3, fm_type=BasisSet.FOURIER, reupload_scaling = ReuploadScaling.TOWER)
    print(f"{fm = }")
    ```
    """

    # Process input
    if support is None:
        support = tuple(range(n_qubits))
    elif len(support) != n_qubits:
        raise ValueError("Wrong qubit support supplied")

    if op not in ROTATIONS:
        raise ValueError(
            f"Operation {op} not supported. "
            f"Please provide one from {[rot.__name__ for rot in ROTATIONS]}."
        )

    scaled_fparam = fm_parameter_scaling(
        fm_type, param, feature_range=feature_range, target_range=target_range
    )

    transform_func = fm_parameter_func(fm_type)

    basis_tag = fm_type.value if isinstance(fm_type, BasisSet) else str(fm_type)
    rs_func, rs_tag = fm_reupload_scaling_fn(reupload_scaling)

    # Set overall multiplier
    multiplier = 1 if multiplier is None else Parameter(multiplier)

    # Build feature map
    op_list = []
    fparam = scaled_fparam
    for i, qubit in enumerate(support):
        if param_prefix is not None:
            train_param = VariationalParameter(param_prefix + f"_{i}")
            fparam = train_param * scaled_fparam
        op_list.append(op(qubit, multiplier * rs_func(i) * transform_func(fparam)))
    fm = kron(*op_list)

    fm.tag = rs_tag + " " + basis_tag + " FM"

    return fm


def exp_fourier_feature_map(
    n_qubits: int,
    support: Optional[tuple[int, ...]] = None,
    param: str = "x",
    feature_range: Optional[tuple[float, float]] = None,
) -> AbstractBlock:
    """
    Exponential fourier feature map.

    Args:
        n_qubits: number of qubits in the feature
        support: qubit support
        param: name of feature `Parameter`
        feature_range: min and max value of the feature, as floats in a Tuple
    """

    if feature_range is None:
        feature_range = (0.0, 2.0**n_qubits)

    support = tuple(range(n_qubits)) if support is None else support
    hlayer = kron(H(qubit) for qubit in support)
    rlayer = feature_map(
        n_qubits,
        support=support,
        param=param,
        op=RZ,
        fm_type=BasisSet.FOURIER,
        reupload_scaling=ReuploadScaling.EXP,
        feature_range=feature_range,
        target_range=(0.0, 2 * PI),
    )
    rlayer.tag = None
    return tag(chain(hlayer, rlayer), f"ExpFourierFM({param})")



from __future__ import annotations

from dataclasses import dataclass, field, fields
from logging import getLogger
from typing import Callable, Type

from sympy import Basic

from qadence.blocks.analog import AnalogBlock
from qadence.blocks.primitive import ParametricBlock
from qadence.operations import RX, AnalogRX
from qadence.parameters import Parameter
from qadence.types import (
    AnsatzType,
    BasisSet,
    MultivariateStrategy,
    ReuploadScaling,
    Strategy,
)

logger = getLogger(__file__)


@dataclass
class FeatureMapConfig:
    num_features: int = 0
    """
    Number of feature parameters to be encoded.

    Defaults to 0. Thus, no feature parameters are encoded.
    """

    basis_set: BasisSet | dict[str, BasisSet] = BasisSet.FOURIER
    """
    Basis set for feature encoding.

    Takes qadence.BasisSet.
    Give a single BasisSet to use the same for all features.
    Give a dict of (str, BasisSet) where the key is the name of the variable and the
    value is the BasisSet to use for encoding that feature.
    BasisSet.FOURIER for Fourier encoding.
    BasisSet.CHEBYSHEV for Chebyshev encoding.
    """

    reupload_scaling: ReuploadScaling | dict[str, ReuploadScaling] = ReuploadScaling.CONSTANT
    """
    Scaling for encoding the same feature on different qubits.

    Scaling used to encode the same feature on different qubits in the
    same layer of the feature maps. Takes qadence.ReuploadScaling.
    Give a single ReuploadScaling to use the same for all features.
    Give a dict of (str, ReuploadScaling) where the key is the name of the variable and the
    value is the ReuploadScaling to use for encoding that feature.
    ReuploadScaling.CONSTANT for constant scaling.
    ReuploadScaling.TOWER for linearly increasing scaling.
    ReuploadScaling.EXP for exponentially increasing scaling.
    """

    feature_range: tuple[float, float] | dict[str, tuple[float, float]] | None = None
    """
    Range of data that the input data is assumed to come from.

    Give a single tuple to use the same range for all features.
    Give a dict of (str, tuple) where the key is the name of the variable and the
    value is the feature range to use for that feature.
    """

    target_range: tuple[float, float] | dict[str, tuple[float, float]] | None = None
    """
    Range of data the data encoder assumes as natural range.

    Give a single tuple to use the same range for all features.
    Give a dict of (str, tuple) where the key is the name of the variable and the
    value is the target range to use for that feature.
    """

    multivariate_strategy: MultivariateStrategy = MultivariateStrategy.PARALLEL
    """
    The encoding strategy in case of multi-variate function.

    Takes qadence.MultivariateStrategy.
    If PARALLEL, the features are encoded in one block of rotation gates
    with the register being split in sub-registers for each feature.
    If SERIES, the features are encoded sequentially using the full register for each feature, with
    an ansatz block between them. PARALLEL is allowed only for DIGITAL `feature_map_strategy`.
    """

    feature_map_strategy: Strategy = Strategy.DIGITAL
    """
    Strategy for feature map.

    Accepts DIGITAL, ANALOG or RYDBERG. Defaults to DIGITAL.
    If the strategy is incompatible with the `operation` chosen, then `operation`
    gets preference and the given strategy is ignored.
    """

    param_prefix: str | None = None
    """
    String prefix to create trainable parameters in Feature Map.

    A string prefix to create trainable parameters multiplying the feature parameter
    inside the feature-encoding function. Note that currently this does not take into
    account the domain of the feature-encoding function.
    Defaults to `None` and thus, the feature map is not trainable.
    Note that this is separate from the name of the parameter.
    The user can provide a single prefix for all features, and it will be appended
    by appropriate feature name automatically.
    """

    num_repeats: int | dict[str, int] = 0
    """
    Number of feature map layers repeated in the data reuploading step.

    If all features are to be repeated the same number of times, then can give a single
    `int`. For different number of repetitions for each feature, provide a dict
    of (str, int) where the key is the name of the variable and the value is the
    number of repetitions for that feature.
    This amounts to the number of additional reuploads. So if `num_repeats` is N,
    the data gets uploaded N+1 times. Defaults to no repetition.
    """

    operation: Callable[[Parameter | Basic], AnalogBlock] | Type[RX] | None = None
    """
    Type of operation.

    Choose among the analog or digital rotations or a custom
    callable function returning an AnalogBlock instance. If the type of operation is
    incompatible with the `strategy` chosen, then `operation` gets preference and
    the given strategy is ignored.
    """

    inputs: list[Basic | str] | None = None
    """
    List that indicates the order of variables of the tensors that are passed.

    Optional if a single feature is being encoded, required otherwise. Given input tensors
    `xs = torch.rand(batch_size, input_size:=2)` a QNN with `inputs=["t", "x"]` will
    assign `t, x = xs[:,0], xs[:,1]`.
    """

    tag: str | None = None
    """
    String to indicate the name tag of the feature map.

    Defaults to None, in which case no tag will be applied.
    """

    def __post_init__(self) -> None:
        if self.multivariate_strategy == MultivariateStrategy.PARALLEL and self.num_features > 1:
            assert (
                self.feature_map_strategy == Strategy.DIGITAL
            ), "For parallel encoding of multiple features, the `feature_map_strategy` must be \
                  of `Strategy.DIGITAL`."

        if self.operation is None:
            if self.feature_map_strategy == Strategy.DIGITAL:
                self.operation = RX
            elif self.feature_map_strategy == Strategy.ANALOG:
                self.operation = AnalogRX  # type: ignore[assignment]

        else:
            if self.feature_map_strategy == Strategy.DIGITAL:
                if isinstance(self.operation, AnalogBlock):
                    logger.warning(
                        "The `operation` is of type `AnalogBlock` but the `feature_map_strategy` is\
                        `Strategy.DIGITAL`. The `feature_map_strategy` will be modified and given \
                        operation will be used."
                    )

                    self.feature_map_strategy = Strategy.ANALOG

            elif self.feature_map_strategy == Strategy.ANALOG:
                if isinstance(self.operation, ParametricBlock):
                    logger.warning(
                        "The `operation` is a digital gate but the `feature_map_strategy` is\
                        `Strategy.ANALOG`. The `feature_map_strategy` will be modified and given\
                        operation will be used."
                    )

                    self.feature_map_strategy = Strategy.DIGITAL

            elif self.feature_map_strategy == Strategy.RYDBERG:
                if self.operation is not None:
                    logger.warning(
                        f"feature_map_strategy is `Strategy.RYDBERG` which does not take any\
                        operation. But an operation {self.operation} is provided. The \
                        `feature_map_strategy` will be modified and given operation will be used."
                    )

                    if isinstance(self.operation, AnalogBlock):
                        self.feature_map_strategy = Strategy.ANALOG
                    else:
                        self.feature_map_strategy = Strategy.DIGITAL

        if self.inputs is not None:
            assert (
                len(self.inputs) == self.num_features
            ), "Inputs list must be of same size as the number of features"
        else:
            if self.num_features == 0:
                self.inputs = []
            elif self.num_features == 1:
                self.inputs = ["x"]
            else:
                raise ValueError(
                    """
                    Your QNN has more than one input. Please provide a list of inputs in the order
                    of your tensor domain. For example, if you want to pass
                    `xs = torch.rand(batch_size, input_size:=3)` to you QNN, where
                    ```
                    t = x[:,0]
                    x = x[:,1]
                    y = x[:,2]
                    ```
                    you have to specify
                    ```
                    inputs=["t", "x", "y"]
                    ```
                    You can also pass a list of sympy symbols.
                """
                )

        property_list = [
            "basis_set",
            "reupload_scaling",
            "feature_range",
            "target_range",
            "num_repeats",
        ]

        for target_field in fields(self):
            if target_field.name in property_list:
                prop = getattr(self, target_field.name)
                if isinstance(prop, dict):
                    assert set(prop.keys()) == set(
                        self.inputs
                    ), f"The keys in {target_field.name} must be the same as the inputs provided. \
                    Alternatively, provide a single value of {target_field.name} to use the same\
                    {target_field.name} for all features."
                else:
                    prop = {key: prop for key in self.inputs}
                    setattr(self, target_field.name, prop)


@dataclass
class AnsatzConfig:
    depth: int = 1
    """Number of layers of the ansatz."""

    ansatz_type: AnsatzType = AnsatzType.HEA
    """What type of ansatz.

    `AnsatzType.HEA` for Hardware Efficient Ansatz.
    `AnsatzType.IIA` for Identity Intialized Ansatz.
    `AnsatzType.ALA` for Alternating Layer Ansatz.
    """

    ansatz_strategy: Strategy = Strategy.DIGITAL
    """Ansatz strategy.

    `Strategy.DIGITAL` for fully digital ansatz. Required if `ansatz_type` is `AnsatzType.ALA`.
    `Strategy.SDAQC` for analog entangling block. Only available for `AnsatzType.HEA` or
    `AnsatzType.ALA`.
    `Strategy.RYDBERG` for fully rydberg hea ansatz. Only available for `AnsatzType.HEA`.
    """

    strategy_args: dict = field(default_factory=dict)
    """
    A dictionary containing keyword arguments to the function creating the ansatz.

    Details about each below.

    For `Strategy.DIGITAL` strategy, accepts the following:
        periodic (bool): if the qubits should be linked periodically.
            periodic=False is not supported in emu-c.
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
            Defaults to  [RX, RY, RX] for hea and [RX, RY] for iia.
        entangler (AbstractBlock): 2-qubit entangling operation.
            Supports CNOT, CZ, CRX, CRY, CRZ, CPHASE. Controlld rotations
            will have variational parameters on the rotation angles.
            Defaults to CNOT

    For `Strategy.SDAQC` strategy, accepts the following:
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
            Defaults to  [RX, RY, RX] for hea and [RX, RY] for iia.
        entangler (AbstractBlock): Hamiltonian generator for the
            analog entangling layer. Time parameter is considered variational.
            Defaults to NN interaction.

    For `Strategy.RYDBERG` strategy, accepts the following:
        addressable_detuning: whether to turn on the trainable semi-local addressing pattern
            on the detuning (n_i terms in the Hamiltonian).
            Defaults to True.
        addressable_drive: whether to turn on the trainable semi-local addressing pattern
            on the drive (sigma_i^x terms in the Hamiltonian).
            Defaults to False.
        tunable_phase: whether to have a tunable phase to get both sigma^x and sigma^y rotations
            in the drive term. If False, only a sigma^x term will be included in the drive part
            of the Hamiltonian generator.
            Defaults to False.
    """
    # The default for a dataclass can not be a mutable object without using this default_factory.

    m_block_qubits: int | None = None
    """
    The number of qubits in the local entangling block of an Alternating Layer Ansatz (ALA).

    Only used when `ansatz_type` is `AnsatzType.ALA`.
    """

    param_prefix: str = "theta"
    """The base bame of the variational parameter."""

    tag: str | None = None
    """
    String to indicate the name tag of the ansatz.

    Defaults to None, in which case no tag will be applied.
    """

    def __post_init__(self) -> None:
        if self.ansatz_type == AnsatzType.IIA:
            assert (
                self.ansatz_strategy != Strategy.RYDBERG
            ), "Rydberg strategy not allowed for Identity-initialized ansatz."

        if self.ansatz_type == AnsatzType.ALA:
            assert (
                self.ansatz_strategy == Strategy.DIGITAL
            ), f"{self.ansatz_strategy} not allowed for Alternating Layer Ansatz.\
            Only `Strategy.DIGITAL` allowed."

            assert (
                self.m_block_qubits is not None
            ), "m_block_qubits must be specified for Alternating Layer Ansatz."



from __future__ import annotations

import itertools
from typing import Any, Type, Union

from qadence.blocks import AbstractBlock, block_is_qubit_hamiltonian, chain, kron, tag
from qadence.constructors.hamiltonians import hamiltonian_factory
from qadence.operations import CNOT, CPHASE, CRX, CRY, CRZ, CZ, RX, RY, HamEvo
from qadence.types import Interaction, Strategy
from typing import Optional


DigitalEntanglers = Union[CNOT, CZ, CRZ, CRY, CRX]


def hea(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "theta",
    support: Optional[tuple[int, ...]] = None,
    strategy: Strategy = Strategy.DIGITAL,
    **strategy_args: Any,
) -> AbstractBlock:
    """
    Factory function for the Hardware Efficient Ansatz (HEA).

    Args:
        n_qubits: number of qubits in the block
        depth: number of layers of the HEA
        param_prefix: the base name of the variational parameters
        support: qubit indexes where the HEA is applied
        strategy: Strategy.Digital or Strategy.DigitalAnalog
        **strategy_args: see below

    Keyword Arguments:
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer. Valid for
            Digital and DigitalAnalog HEA.
        periodic (bool): if the qubits should be linked periodically.
            periodic=False is not supported in emu-c. Valid for only
            for Digital HEA.
        entangler (AbstractBlock):
            - Digital: 2-qubit entangling operation. Supports CNOT, CZ,
            CRX, CRY, CRZ, CPHASE. Controlled rotations will have variational
            parameters on the rotation angles.
            - DigitaAnalog | Analog: Hamiltonian generator for the
            analog entangling layer. Defaults to global ZZ Hamiltonian.
            Time parameter is considered variational.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence import RZ, RX
    from qadence_gradient import hea

    # create the circuit
    n_qubits, depth = 2, 4
    ansatz = hea(
        n_qubits=n_qubits,
        depth=depth,
        strategy="sDAQC",
        operations=[RZ,RX,RZ]
    )
    ```
    """

    if support is None:
        support = tuple(range(n_qubits))

    hea_func_dict = {
        Strategy.DIGITAL: hea_digital,
        Strategy.SDAQC: hea_sDAQC,
        Strategy.BDAQC: hea_bDAQC,
        Strategy.ANALOG: hea_analog,
    }

    try:
        hea_func = hea_func_dict[strategy]
    except KeyError:
        raise KeyError(f"Strategy {strategy} not recognized.")

    hea_block: AbstractBlock = hea_func(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        **strategy_args,
    )  # type: ignore

    return hea_block


#############
## DIGITAL ##
#############


def _rotations_digital(
    n_qubits: int,
    depth: int,
    param_prefix: str = "theta",
    support: Optional[tuple[int, ...]] = None,
    operations: list[Type[AbstractBlock]] = [RX, RY, RX],
) -> list[AbstractBlock]:
    """Creates the layers of single qubit rotations in an HEA."""
    if support is None:
        support = tuple(range(n_qubits))
    iterator = itertools.count()
    rot_list: list[AbstractBlock] = []
    for d in range(depth):
        rots = [
            kron(
                gate(support[n], param_prefix + f"_{next(iterator)}")  # type: ignore [arg-type]
                for n in range(n_qubits)
            )
            for gate in operations
        ]
        rot_list.append(chain(*rots))
    return rot_list


def _entangler(
    control: int,
    target: int,
    param_str: str,
    op: Type[DigitalEntanglers] = CNOT,
) -> AbstractBlock:
    if op in [CNOT, CZ]:
        return op(control, target)  # type: ignore
    elif op in [CRZ, CRY, CRX, CPHASE]:
        return op(control, target, param_str)  # type: ignore
    else:
        raise ValueError("Provided entangler not accepted for digital HEA.")


def _entanglers_digital(
    n_qubits: int,
    depth: int,
    param_prefix: str = "theta",
    support: Optional[tuple[int, ...]] = None,
    periodic: bool = False,
    entangler: Type[DigitalEntanglers] = CNOT,
) -> list[AbstractBlock]:
    """Creates the layers of digital entangling operations in an HEA."""
    if support is None:
        support = tuple(range(n_qubits))
    iterator = itertools.count()
    ent_list: list[AbstractBlock] = []
    for d in range(depth):
        ents = []
        ents.append(
            kron(
                _entangler(
                    control=support[n],
                    target=support[n + 1],
                    param_str=param_prefix + f"_ent_{next(iterator)}",
                    op=entangler,
                )
                for n in range(n_qubits)
                if not n % 2 and n < n_qubits - 1
            )
        )
        if n_qubits > 2:
            ents.append(
                kron(
                    _entangler(
                        control=support[n],
                        target=support[(n + 1) % n_qubits],
                        param_str=param_prefix + f"_ent_{next(iterator)}",
                        op=entangler,
                    )
                    for n in range(n_qubits - (not periodic))
                    if n % 2
                )
            )
        ent_list.append(chain(*ents))
    return ent_list


def hea_digital(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "theta",
    periodic: bool = False,
    operations: list[type[AbstractBlock]] = [RX, RY, RX],
    support: Optional[tuple[int, ...]] = None,
    entangler: Type[DigitalEntanglers] = CNOT,
) -> AbstractBlock:
    """
    Construct the Digital Hardware Efficient Ansatz (HEA).

    Args:
        n_qubits (int): number of qubits in the block.
        depth (int): number of layers of the HEA.
        param_prefix (str): the base name of the variational parameters
        periodic (bool): if the qubits should be linked periodically.
            periodic=False is not supported in emu-c.
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
        support (tuple): qubit indexes where the HEA is applied.
        entangler (AbstractBlock): 2-qubit entangling operation.
            Supports CNOT, CZ, CRX, CRY, CRZ. Controlld rotations
            will have variational parameters on the rotation angles.
    """
    try:
        if entangler not in [CNOT, CZ, CRX, CRY, CRZ, CPHASE]:
            raise ValueError(
                "Please provide a valid two-qubit entangler operation for digital HEA."
            )
    except TypeError:
        raise ValueError("Please provide a valid two-qubit entangler operation for digital HEA.")

    rot_list = _rotations_digital(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        operations=operations,
    )

    ent_list = _entanglers_digital(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        periodic=periodic,
        entangler=entangler,
    )

    layers = []
    for d in range(depth):
        layers.append(rot_list[d])
        layers.append(ent_list[d])
    return tag(chain(*layers), "HEA")


###########
## sDAQC ##
###########


def _entanglers_analog(
    depth: int,
    param_prefix: str = "theta",
    entangler: AbstractBlock | None = None,
) -> list[AbstractBlock]:
    return [HamEvo(entangler, param_prefix + f"_t_{d}") for d in range(depth)]


def hea_sDAQC(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "theta",
    operations: list[type[AbstractBlock]] = [RX, RY, RX],
    support: Optional[tuple[int, ...]] = None,
    entangler: AbstractBlock | None = None,
) -> AbstractBlock:
    """
    Construct the Hardware Efficient Ansatz (HEA) with analog entangling layers.

    It uses step-wise digital-analog computation.

    Args:
        n_qubits (int): number of qubits in the block.
        depth (int): number of layers of the HEA.
        param_prefix (str): the base name of the variational parameters
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
        support (tuple): qubit indexes where the HEA is applied.
        entangler (AbstractBlock): Hamiltonian generator for the
            analog entangling layer. Defaults to global ZZ Hamiltonian.
            Time parameter is considered variational.
    """

    # TODO: Add qubit support
    if entangler is None:
        entangler = hamiltonian_factory(n_qubits, interaction=Interaction.NN)
    try:
        if not block_is_qubit_hamiltonian(entangler):
            raise ValueError(
                "Please provide a valid Pauli Hamiltonian generator for digital-analog HEA."
            )
    except NotImplementedError:
        raise ValueError(
            "Please provide a valid Pauli Hamiltonian generator for digital-analog HEA."
        )

    rot_list = _rotations_digital(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        operations=operations,
    )

    ent_list = _entanglers_analog(
        depth=depth,
        param_prefix=param_prefix,
        entangler=entangler,
    )

    layers = []
    for d in range(depth):
        layers.append(rot_list[d])
        layers.append(ent_list[d])
    return tag(chain(*layers), "HEA-sDA")


###########
## bDAQC ##
###########


def hea_bDAQC(*args: Any, **kwargs: Any) -> Any:
    raise NotImplementedError


############
## ANALOG ##
############


def hea_analog(*args: Any, **kwargs: Any) -> Any:
    raise NotImplementedError



from __future__ import annotations

from functools import singledispatch
from typing import Any

from torch import Tensor, rand

from qadence.blocks import AbstractBlock, parameters
from qadence.circuit import QuantumCircuit
from qadence.parameters import Parameter, stringify
from qadence.model import QuantumModel
from .qnn_model import QNN


@singledispatch
def rand_featureparameters(
    x: QuantumCircuit | AbstractBlock | QuantumModel | QNN, *args: Any
) -> dict[str, Tensor]:
    raise NotImplementedError(f"Unable to generate random featureparameters for object {type(x)}.")


@rand_featureparameters.register
def _(block: AbstractBlock, batch_size: int = 1) -> dict[str, Tensor]:
    non_number_params = [p for p in parameters(block) if not p.is_number]
    feat_params: list[Parameter] = [p for p in non_number_params if not p.trainable]
    return {stringify(p): rand(batch_size, requires_grad=False) for p in feat_params}


@rand_featureparameters.register
def _(circuit: QuantumCircuit, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(circuit.block, batch_size)


@rand_featureparameters.register
def _(qm: QuantumModel, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(qm._circuit.abstract, batch_size)


@rand_featureparameters.register
def _(qnn: QNN, batch_size: int = 1) -> dict[str, Tensor]:
    return rand_featureparameters(qnn._circuit.abstract, batch_size)



from __future__ import annotations

import torch
from torch import Tensor
from torch.autograd import grad


def hessian(output: Tensor, inputs: list[Tensor]) -> Tensor:
    """Calculates the Hessian of a given output vector wrt the inputs.

    TODO: Use autograd built-in functions for a more efficient implementation,
    but grad tree
    is broken by the Overlap method

    Args:
        output (Tensor): Output vector
        inputs (list[Tensor]): List of input parameters
    """

    jacobian = grad(
        output,
        inputs,
        torch.ones_like(output),
        create_graph=True,
        allow_unused=True,
    )

    n_params = len(inputs)
    hess = torch.empty((n_params, n_params))
    for i in range(n_params):
        jacobian_var = jacobian[i]
        ovrlp_grad2 = grad(
            jacobian_var,
            inputs,
            torch.ones_like(jacobian_var),
            create_graph=True,
            allow_unused=True,
        )
        for j in range(n_params):
            hess[i, j] = ovrlp_grad2[j]

    return hess



from __future__ import annotations

from typing import Any, Callable
import numpy as np
from sympy import Basic

from qadence.backend import BackendConfiguration
from qadence.blocks import chain, kron
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.composite import ChainBlock, KronBlock
from qadence.blocks.utils import tag, add
from qadence.parameters import Parameter
from qadence.circuit import QuantumCircuit
from qadence.constructors import (
    analog_feature_map,
    feature_map,
    hamiltonian_factory,
    rydberg_feature_map,
    rydberg_hea,
    rydberg_tower_feature_map,
)
from qadence.constructors.ala import ala_digital
from qadence.constructors.hamiltonians import ObservableConfig, TDetuning
from qadence.constructors.hea import hea_digital, hea_sDAQC
from qadence.constructors.iia import iia
from qadence.measurements import Measurements
from qadence.noise import NoiseHandler
from qadence.operations import CNOT, RX, RY, RZ, I
from qadence.register import Register
from qadence.types import (
    AnsatzType,
    BackendName,
    DiffMode,
    InputDiffMode,
    Interaction,
    MultivariateStrategy,
    ReuploadScaling,
    Strategy,
    TParameter,
    TArray,
)

from .qnn_config import AnsatzConfig, FeatureMapConfig
from .qnn_model import QNN


def _create_support_arrays(
    num_qubits: int,
    num_features: int,
    multivariate_strategy: MultivariateStrategy,
) -> list[tuple[int, ...]]:
    """
    Create the support arrays for the digital feature map.

    Args:
        num_qubits (int): The number of qubits.
        num_features (int): The number of features.
        multivariate_strategy (MultivariateStrategy): The multivariate encoding strategy.
            Either 'MultivariateStrategy.SERIES' or 'MultivariateStrategy.PARALLEL'.

    Returns:
        list[tuple[int, ...]]: The list of support arrays. ith element of the list is the support
            array for the ith feature.

    Raises:
        ValueError: If the number of features is greater than the number of qubits
            and the strategy is `MultivariateStrategy.PARALLEL` not possible to assign a support
            array to each feature.
        ValueError: If the multivariate strategy is not `MultivariateStrategy.SERIES` or
            `MultivariateStrategy.PARALLEL`.
    """
    if multivariate_strategy == MultivariateStrategy.SERIES:
        return [tuple(range(num_qubits)) for i in range(num_features)]
    elif multivariate_strategy == MultivariateStrategy.PARALLEL:
        if num_features <= num_qubits:
            return [tuple(x.tolist()) for x in np.array_split(np.arange(num_qubits), num_features)]
        else:
            raise ValueError(
                f"Number of features {num_features} must be less than or equal to the number of \
                qubits {num_qubits} if the features are to be encoded parallely."
            )
    else:
        raise ValueError(
            f"Invalid encoding strategy {multivariate_strategy} provided. Only \
                `MultivariateStrategy.SERIES` or `MultivariateStrategy.PARALLEL` are allowed."
        )


def _encode_features_series_digital(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Encode the features in series using digital feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of digital feature map blocks.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    support_arrays_list = _create_support_arrays(
        num_qubits=num_qubits,
        num_features=config.num_features,
        multivariate_strategy=config.multivariate_strategy,
    )

    support_arrays = {
        key: support
        for key, support in zip(config.inputs, support_arrays_list)  # type: ignore[union-attr, arg-type]
    }

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    if config.param_prefix is None:
        param_prefixes = {feature: None for feature in config.inputs}  # type: ignore[union-attr]
    else:
        param_prefixes = {feature: f"{config.param_prefix}_{feature}" for feature in config.inputs}  # type: ignore

    fm_blocks: list[AbstractBlock] = []

    for i in range(max(num_uploads.values())):
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_blocks.append(
                    feature_map(
                        num_qubits,
                        support=support_arrays[feature],
                        param=feature,
                        op=config.operation,  # type: ignore[arg-type]
                        fm_type=config.basis_set[feature],  # type: ignore[arg-type, index]
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[arg-type, index]
                        feature_range=config.feature_range[feature],  # type: ignore[arg-type, index]
                        target_range=config.target_range[feature],  # type: ignore[arg-type, index]
                        param_prefix=param_prefixes[feature],
                    )
                )
            num_uploads[feature] -= 1

    return fm_blocks


def _encode_features_parallel_digital(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Encode the features in parallel using digital feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of digital feature map blocks.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    support_arrays_list = _create_support_arrays(
        num_qubits=num_qubits,
        num_features=config.num_features,
        multivariate_strategy=config.multivariate_strategy,
    )

    support_arrays = {
        key: support
        for key, support in zip(config.inputs, support_arrays_list)  # type: ignore[union-attr, arg-type]
    }

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    if config.param_prefix is None:
        param_prefixes = {feature: None for feature in config.inputs}  # type: ignore[union-attr]
    else:
        param_prefixes = {feature: f"{config.param_prefix}_{feature}" for feature in config.inputs}  # type: ignore

    fm_blocks: list[AbstractBlock] = []

    for i in range(max(num_uploads.values())):
        fm_layer = []
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_layer.append(
                    feature_map(
                        len(support_arrays[feature]),
                        support=support_arrays[feature],
                        param=feature,
                        op=config.operation,  # type: ignore[arg-type]
                        fm_type=config.basis_set[feature],  # type: ignore[index]
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[index]
                        feature_range=config.feature_range[feature],  # type: ignore[arg-type, index]
                        target_range=config.target_range[feature],  # type: ignore[arg-type, index]
                        param_prefix=param_prefixes[feature],
                    )
                )
            num_uploads[feature] -= 1

        fm_blocks.append(kron(*fm_layer))

    return fm_blocks


def _create_digital_fm(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create the digital feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of digital feature map blocks.

    Raises:
        ValueError: If the encoding strategy is invalid. Only `MultivariateStrategy.SERIES` or
            `MultivariateStrategy.PARALLEL` are allowed.
    """
    if config.multivariate_strategy == MultivariateStrategy.SERIES:
        fm_blocks = _encode_features_series_digital(register, config)
    elif config.multivariate_strategy == MultivariateStrategy.PARALLEL:
        fm_blocks = _encode_features_parallel_digital(register, config)
    else:
        raise ValueError(
            f"Invalid encoding strategy {config.multivariate_strategy} provided. Only\
            `MultivariateStrategy.SERIES` or `MultivariateStrategy.PARALLEL` are allowed."
        )

    return fm_blocks


def _create_analog_fm(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create the analog feature map.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (FeatureMapConfig): The feature map configuration.

    Returns:
        list[AbstractBlock]: The list of analog feature map blocks.
    """

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    fm_blocks: list[AbstractBlock] = []

    for i in range(max(num_uploads.values())):
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_blocks.append(
                    analog_feature_map(
                        param=feature,
                        op=config.operation,  # type: ignore[arg-type]
                        fm_type=config.basis_set[feature],  # type: ignore[arg-type, index]
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[arg-type, index]
                        feature_range=config.feature_range[feature],  # type: ignore[arg-type, index]
                        target_range=config.target_range[feature],  # type: ignore[arg-type, index]
                    )
                )
            num_uploads[feature] -= 1

    return fm_blocks


def _encode_feature_rydberg(
    num_qubits: int,
    param: str,
    reupload_scaling: ReuploadScaling,
) -> AbstractBlock:
    """
    Encode features using a Rydberg feature map.

    Args:
        num_qubits (int): The number of qubits to encode the features on.
        param (str): The parameter prefix to use for the feature map parameter names.
            reupload_scaling (ReuploadScaling): The scaling strategy for reuploads.

    Returns:
        The Rydberg feature map.

    Raises:
        NotImplementedError: If the reupload scaling strategy is not implemented.
            Only `ReuploadScaling.CONSTANT` and `ReuploadScaling.TOWER` are supported.
    """
    if reupload_scaling == ReuploadScaling.CONSTANT:
        return rydberg_feature_map(n_qubits=num_qubits, param=param)

    elif reupload_scaling == ReuploadScaling.TOWER:
        return rydberg_tower_feature_map(n_qubits=num_qubits, param=param)

    else:
        raise NotImplementedError(f"Rydberg feature map not implemented for {reupload_scaling}")


def _create_rydberg_fm(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create a Rydberg feature map for the given configuration.

    Args:
        register (int | Register): The number of qubits or the register to apply the feature map to.
        config (FeatureMapConfig): The configuration for the feature map.

    Returns:
        list: A list of Rydberg feature map blocks.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    num_uploads = {key: value + 1 for key, value in config.num_repeats.items()}  # type: ignore[union-attr]

    fm_blocks = []

    for i in range(max(num_uploads.values())):
        for feature in config.inputs:  # type: ignore[union-attr]
            if num_uploads[feature] > 0:
                fm_blocks.append(
                    _encode_feature_rydberg(
                        num_qubits=num_qubits,
                        param=feature,
                        reupload_scaling=config.reupload_scaling[feature],  # type: ignore[arg-type, index]
                    )
                )
            num_uploads[feature] -= 1

    return fm_blocks


def create_fm_blocks(
    register: int | Register,
    config: FeatureMapConfig,
) -> list[AbstractBlock]:
    """
    Create a list of feature map blocks based on the given configuration.

    In case of series encoding or even parallel encoding with data reuploads,
    the outputs is a list of blocks that still need to be interleaved with non
    commuting blocks.

    Args:
        register (int | Register): The number of qubits or the register.
        config (FeatureMapConfig): The configuration for the feature map.

    Returns:
        list[AbstractBlock]: A list of feature map blocks.

    Raises:
        ValueError: If the feature map strategy is not `Strategy.DIGITAL`, `Strategy.ANALOG` or
            `Strategy.RYDBERG`.
    """
    if config.feature_map_strategy == Strategy.DIGITAL:
        return _create_digital_fm(register=register, config=config)
    elif config.feature_map_strategy == Strategy.ANALOG:
        return _create_analog_fm(register=register, config=config)
    elif config.feature_map_strategy == Strategy.RYDBERG:
        return _create_rydberg_fm(register=register, config=config)
    else:
        raise NotImplementedError(
            f"Feature map not implemented for strategy {config.feature_map_strategy}. \
            Only `Strategy.DIGITAL`, `Strategy.ANALOG` or `Strategy.RYDBERG` allowed."
        )


def _ansatz_layer(
    register: int | Register,
    ansatz_config: AnsatzConfig,
    index: int,
) -> AbstractBlock:
    """
    Create a layer of the ansatz based on the configuration.

    Args:
        register (int | Register): The number of qubits or the register.
        ansatz_config (AnsatzConfig): The configuration for the ansatz.
        index (int): The index of the layer.

    Returns:
        AbstractBlock: The layer of the ansatz.
    """
    new_config = AnsatzConfig(
        depth=1,
        ansatz_type=ansatz_config.ansatz_type,
        ansatz_strategy=ansatz_config.ansatz_strategy,
        strategy_args=ansatz_config.strategy_args,
        param_prefix=f"fm_{index}",
    )

    return create_ansatz(register=register, config=new_config)


def _create_iia_digital(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Digital Identity Initialized Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Identity Initialized Ansatz.
    """
    operations = config.strategy_args.get("operations", [RX, RY])
    entangler = config.strategy_args.get("entangler", CNOT)
    periodic = config.strategy_args.get("periodic", False)

    return iia(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        strategy=Strategy.DIGITAL,
        rotations=operations,
        entangler=entangler,
        periodic=periodic,
    )


def _create_iia_sdaqc(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the SDAQC Identity Initialized Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The SDAQC Identity Initialized Ansatz.
    """
    operations = config.strategy_args.get("operations", [RX, RY])
    entangler = config.strategy_args.get("entangler", CNOT)
    periodic = config.strategy_args.get("periodic", False)

    return iia(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        strategy=Strategy.SDAQC,
        rotations=operations,
        entangler=entangler,
        periodic=periodic,
    )


def _create_iia(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Identity Initialized Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Identity Initialized Ansatz.

    Raises:
        ValueError: If the ansatz strategy is not supported. Only `Strategy.DIGITAL` and
            `Strategy.SDAQC` are allowed.
    """
    if config.ansatz_strategy == Strategy.DIGITAL:
        return _create_iia_digital(num_qubits=num_qubits, config=config)
    elif config.ansatz_strategy == Strategy.SDAQC:
        return _create_iia_sdaqc(num_qubits=num_qubits, config=config)
    else:
        raise ValueError(
            f"Invalid ansatz strategy {config.ansatz_strategy} provided. Only `Strategy.DIGITAL` \
                and `Strategy.SDAQC` allowed for IIA."
        )


def _create_hea_digital(num_qubits: int, config: AnsatzConfig) -> AbstractBlock:
    """
    Create the Digital Hardware Efficient Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Digital Hardware Efficient Ansatz.
    """
    operations = config.strategy_args.get("operations", [RX, RY, RX])
    entangler = config.strategy_args.get("entangler", CNOT)
    periodic = config.strategy_args.get("periodic", False)

    return hea_digital(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        operations=operations,
        entangler=entangler,
        periodic=periodic,
    )


def _create_hea_sdaqc(num_qubits: int, config: AnsatzConfig) -> AbstractBlock:
    """
    Create the SDAQC Hardware Efficient Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The SDAQC Hardware Efficient Ansatz.
    """
    operations = config.strategy_args.get("operations", [RX, RY, RX])
    entangler = config.strategy_args.get(
        "entangler", hamiltonian_factory(num_qubits, interaction=Interaction.NN)
    )

    return hea_sDAQC(
        n_qubits=num_qubits,
        depth=config.depth,
        param_prefix=config.param_prefix,
        operations=operations,
        entangler=entangler,
    )


def _create_hea_rydberg(
    register: int | Register,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Rydberg Hardware Efficient Ansatz based on the configuration.

    Args:
        register (int | Register): The number of qubits or the register object.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Rydberg Hardware Efficient Ansatz.
    """
    register = register if isinstance(register, Register) else Register.circle(n_qubits=register)

    addressable_detuning = config.strategy_args.get("addressable_detuning", True)
    addressable_drive = config.strategy_args.get("addressable_drive", False)
    tunable_phase = config.strategy_args.get("tunable_phase", False)

    return rydberg_hea(
        register=register,
        n_layers=config.depth,
        addressable_detuning=addressable_detuning,
        addressable_drive=addressable_drive,
        tunable_phase=tunable_phase,
        additional_prefix=config.param_prefix,
    )


def _create_hea(
    register: int | Register,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Hardware Efficient Ansatz based on the configuration.

    Args:
        register (int | Register): The number of qubits or the register to create the ansatz for.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The hardware efficient ansatz block.

    Raises:
        ValueError: If the ansatz strategy is not `Strategy.DIGITAL`, `Strategy.SDAQC`, or
            `Strategy.RYDBERG`.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    if config.ansatz_strategy == Strategy.DIGITAL:
        return _create_hea_digital(num_qubits=num_qubits, config=config)
    elif config.ansatz_strategy == Strategy.SDAQC:
        return _create_hea_sdaqc(num_qubits=num_qubits, config=config)
    elif config.ansatz_strategy == Strategy.RYDBERG:
        return _create_hea_rydberg(register=register, config=config)
    else:
        raise ValueError(
            f"Invalid ansatz strategy {config.ansatz_strategy} provided. Only `Strategy.DIGITAL`, \
                `Strategy.SDAQC`, and `Strategy.RYDBERG` allowed"
        )


def _create_ala_digital(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Digital Alternating Layer Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Digital Alternating Layer Ansatz.
    """
    operations = config.strategy_args.get("operation", [RX, RY, RX])
    entangler = config.strategy_args.get("entangler", CNOT)

    return ala_digital(
        n_qubits=num_qubits,
        m_block_qubits=config.m_block_qubits,  # type: ignore[arg-type]
        param_prefix=config.param_prefix,
        operations=operations,
        entangler=entangler,
    )


def _create_ala(
    num_qubits: int,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the Alternating Layer Ansatz based on the configuration.

    Args:
        num_qubits (int): The number of qubits.
        config (AnsatzConfig): The configuration for the ansatz.

    Returns:
        AbstractBlock: The Alternating Layer Ansatz.

    Raises:
        ValueError: If the ansatz strategy is not `Strategy.DIGITAL`.
    """
    if config.ansatz_strategy == Strategy.DIGITAL:
        return _create_ala_digital(num_qubits=num_qubits, config=config)
    else:
        raise ValueError(
            f"Invalid ansatz strategy {config.ansatz_strategy} provided. Only `Strategy.DIGITAL` \
                allowed"
        )


def create_ansatz(
    register: int | Register,
    config: AnsatzConfig,
) -> AbstractBlock:
    """
    Create the ansatz based on the configuration.

    Args:
        register (int | Register): Number of qubits or a register object.
        config (AnsatzConfig): Configuration for the ansatz.

    Returns:
        AbstractBlock: The ansatz block.

    Raises:
        NotImplementedError: If the ansatz type is not implemented.
    """
    num_qubits = register if isinstance(register, int) else register.n_qubits

    if config.ansatz_type == AnsatzType.IIA:
        return _create_iia(num_qubits=num_qubits, config=config)
    elif config.ansatz_type == AnsatzType.HEA:
        return _create_hea(register=register, config=config)
    elif config.ansatz_type == AnsatzType.ALA:
        return _create_ala(num_qubits=num_qubits, config=config)
    else:
        raise NotImplementedError(
            f"Ansatz of type {config.ansatz_type} not implemented yet. Only `AnsatzType.HEA` and\
                `AnsatzType.IIA` available."
        )


def _interleave_ansatz_in_fm(
    register: int | Register,
    fm_blocks: list[AbstractBlock],
    ansatz_config: AnsatzConfig,
) -> ChainBlock:
    """
    Interleave the ansatz layers in between the feature map layers.

    Args:
        register (int | Register): Number of qubits or a register object.
        fm_blocks (list[AbstractBlock]): List of feature map blocks.
        ansatz_config (AnsatzConfig): Ansatz configuration.

    Returns:
        ChainBlock: A block containing feature map layers with interleaved ansatz layers.
    """
    full_fm = []
    for idx, block in enumerate(fm_blocks):
        full_fm.append(block)
        if idx + 1 < len(fm_blocks):
            full_fm.append(_ansatz_layer(register, ansatz_config, idx))

    return chain(*full_fm)


def _global_identity(register: int | Register) -> KronBlock:
    """Create a global identity block."""
    return kron(
        *[I(i) for i in range(register if isinstance(register, int) else register.n_qubits)]
    )


def create_observable(
    register: int | Register,
    config: ObservableConfig,
) -> AbstractBlock:
    """
    Create an observable block.

    Args:
        register (int | Register): Number of qubits or a register object.
        config (ObservableConfig): Observable configuration.

    Returns:
        AbstractBlock: The observable block.
    """
    shifting_term: AbstractBlock = config.shift * _global_identity(register)  # type: ignore[operator]
    detuning_hamiltonian: AbstractBlock = config.scale * hamiltonian_factory(  # type: ignore[operator]
        register=register,
        interaction=config.interaction,
        detuning=config.detuning,
    )
    obs: AbstractBlock = add(shifting_term, detuning_hamiltonian)

    if isinstance(config.tag, str):
        tag(obs, config.tag)

    return obs


def build_qnn_from_configs(
    register: int | Register,
    observable_config: ObservableConfig | list[ObservableConfig],
    fm_config: FeatureMapConfig = FeatureMapConfig(),
    ansatz_config: AnsatzConfig = AnsatzConfig(),
    backend: BackendName = BackendName.PYQTORCH,
    diff_mode: DiffMode = DiffMode.AD,
    measurement: Measurements | None = None,
    noise: NoiseHandler | None = None,
    configuration: BackendConfiguration | dict | None = None,
    input_diff_mode: InputDiffMode | str = InputDiffMode.AD,
) -> QNN:
    """
    Build a QNN model.

    Args:
        register (int | Register): Number of qubits or a register object.
        observable_config (ObservableConfig | list[ObservableConfig]): Observable configuration(s).
        fm_config (FeatureMapConfig): Feature map configuration.
        ansatz_config (AnsatzConfig): Ansatz configuration.
        backend (BackendName): The chosen quantum backend.
        diff_mode (DiffMode): The differentiation engine to use. Choices are
            'gpsr' or 'ad'.
        measurement (Measurements): Optional measurement protocol. If None,
            use exact expectation value with a statevector simulator.
        noise (Noise): A noise model to use.
        configuration (BackendConfiguration | dict): Optional backend configuration.
        input_diff_mode (InputDiffMode): The differentiation mode for the input tensor.

    Returns:
        QNN: A QNN model.
    """
    blocks: list[AbstractBlock] = []
    inputs: list[Basic | str] | None = None

    if fm_config.num_features > 0:
        fm_blocks = create_fm_blocks(register=register, config=fm_config)
        full_fm = _interleave_ansatz_in_fm(
            register=register,
            fm_blocks=fm_blocks,
            ansatz_config=ansatz_config,
        )
        if isinstance(fm_config.tag, str):
            tag(full_fm, fm_config.tag)
        inputs = fm_config.inputs
        blocks.append(full_fm)

    ansatz = create_ansatz(register=register, config=ansatz_config)
    if isinstance(ansatz_config.tag, str):
        tag(ansatz, ansatz_config.tag)
    blocks.append(ansatz)

    circ = QuantumCircuit(register, *blocks)

    observable: AbstractBlock | list[AbstractBlock] = (
        [create_observable(register=register, config=cfg) for cfg in observable_config]
        if isinstance(observable_config, list)
        else create_observable(register=register, config=observable_config)
    )

    ufa = QNN(
        circ,
        observable,
        inputs=inputs,
        backend=backend,
        diff_mode=diff_mode,
        measurement=measurement,
        noise=noise,
        configuration=configuration,
        input_diff_mode=input_diff_mode,
    )

    return ufa


def _create_feature_map_qcnn(
    n_qubits: int,
    n_inputs: int,
    fm_type: str = "Fourier",
    op: Any = RX,
) -> Any:
    """
    Creates a feature map (FM) by dividing qubits among inputs and applying.

    the specified feature map type.

    Args:
        n_qubits (int): Total number of qubits.
        n_inputs (int): Number of inputs.
        fm_type (str): Type of feature map to use (e.g., "Fourier").
        op (Any): Quantum operation to use in the feature map (e.g., RX).

    Returns:
        Any: The combined feature map as a kronecker product
        of individual feature maps.
    """
    fm_temp = []
    qubits_per_input = n_qubits // n_inputs  # Base number of qubits per input
    exceeding_qubits = n_qubits % n_inputs  # Number of exceeding qubits
    start = 0  # Track current qubit index

    for i in range(n_inputs):
        # Assign base qubits + 1 extra if input has exceeding qubits
        num_qubits = qubits_per_input + 1 if i < exceeding_qubits else qubits_per_input
        end = start + num_qubits

        # Create FM for this input
        fm_temp.append(
            feature_map(
                n_qubits=num_qubits,
                param=f"\u03c6_{i}",  # Use phi_i as the parameter
                op=op,
                fm_type=fm_type,
                support=tuple(range(start, end)),
            )
        )
        start = end  # Update starting index for next FM

    # Combine all feature maps using kronecker product
    return kron(*fm_temp)


def _get_block_params(
    params: dict,
    layer: int,
    rep: int,
    pos: int,
    is_corr: bool = False,
) -> Any:
    """
    Retrieves the parameter for a given operation.

    Args:
        params (dict): Dictionary to store and retrieve parameters.
        layer (int): The index of the current layer.
        rep (int): The index of the current repetition in the layer.
        pos (int): Position of the qubit in the layer.
        is_corr (bool): If True, uses correlated parameters for corresponding gates in W^opt_ij.

    Returns:
        Parameter: the retrieved parameter.
    """
    if is_corr:
        # Cycle pos from 0 to 8
        key = f"θ_{layer}_{pos % 9}"
    else:
        key = f"θ_{layer}_{rep}_{pos}"

    if key not in params:
        params[key] = Parameter(key)
    return params[key]


def _create_single_W(
    params: dict,
    operations: list[Any],
    entangler: Any,
    layer: int,
    rep: int,
    max_reps: int,
    control: int,
    target: int,
    spacing: int,
    n_qubits: int,
    is_corr: bool = False,
) -> ChainBlock:
    """Creates a single convolutional cell W_ij."""
    pad = [
        I(q)
        for q in range(control - spacing, control + spacing + 1)
        if q != control and q != target and 0 <= q < n_qubits
    ]
    gates = []

    # Track per-layer repetition index for proper parameter continuity
    key_param_counter = f"param_index_{layer}_{rep}"
    if key_param_counter not in params:
        params[key_param_counter] = 0  # Initialize if first time

    param_index = params[key_param_counter]  # Load index
    single_params = {}  # Store params for single RZ/RY gates

    # Apply the first sequence of operations
    for _, op in enumerate(operations):
        param_control = _get_block_params(params, layer, rep, param_index, is_corr)
        param_index += 1
        param_target = _get_block_params(params, layer, rep, param_index, is_corr)
        param_index += 1
        gates.append(
            kron(
                *pad,
                op(control, param_control),
                op(target, param_target),
            )
        )
    # entangling gate
    gates.append(entangler(target, control))

    # Apply RZ, RY and entangling gates for intermediate step
    single_params["control_rz"] = _get_block_params(params, layer, rep, param_index, is_corr)
    param_index += 1
    single_params["target_ry"] = _get_block_params(params, layer, rep, param_index, is_corr)
    param_index += 1
    gates.append(
        kron(
            *pad,
            RZ(control, single_params["control_rz"]),
            RY(target, single_params["target_ry"]),
        )
    )
    # entangling gate
    gates.append(entangler(control, target))

    intermediate_ry = _get_block_params(params, layer, rep, param_index, is_corr)
    param_index += 1
    gates.append(
        kron(
            *pad,
            I(control),
            RY(target, intermediate_ry),
        )
    )
    # entangling gate
    gates.append(entangler(target, control))

    # Apply the first sequence of operations
    for _, op in enumerate(operations):
        param_control = _get_block_params(params, layer, rep, param_index, is_corr)
        param_index += 1
        param_target = _get_block_params(params, layer, rep, param_index, is_corr)
        param_index += 1
        gates.append(
            kron(
                *pad,
                op(control, param_control),
                op(target, param_target),
            )
        )
    # Add final entangling gate (control -> target)
    if rep == int(max_reps - 1):
        gates.append(entangler(control, target))

    # Update params dict with the last used index
    params[key_param_counter] = param_index

    return chain(*gates)


def _create_conv_layer(
    layer_index: int,
    max_reps: int,
    current_indices: list[int],
    params: dict,
    operations: list[Any],
    entangler: Any,
    n_qubits: int,
    is_corr: bool,
) -> tuple[AbstractBlock, list[int]]:
    """
    Function to create a single convolutional layer.

    Args:
        layer_index (int): The index of the current layer.
        reps (int): Number of repetitions for this layer.
        current_indices (List[int]): Indices of qubits for the current layer.
        params (dict): Dictionary to store and retrieve parameters.
        operations (List[Any]): List of quantum operations to apply in the gates.
        entangler (Any): Entangling operation, such as CZ.
        n_qubits (int): Total number of qubits.

    Returns:
        Tuple[AbstractBlock, List[int]]: A tuple containing the quantum block
        for the layer and the target indices for the next layer.
    """
    current_layer = []
    next_indices = []
    spacing = layer_index

    if layer_index in [0, 1]:  # Special behavior for first two layers
        layer_reps = []
        for rep in range(max_reps):
            rep_kron = []
            # Define qubit pairs based on odd/even repetition
            if rep % 2 == 0:  # Even d: regular behavior
                pairs = zip(current_indices[::2], current_indices[1::2])
            else:  # Odd d: shift downward, leaving qubits 0 and 7 free
                pairs = zip(current_indices[1:-1:2], current_indices[2:-1:2])

            # Build the gate sequence for each pair
            for control, target in pairs:
                W_pairs = _create_single_W(
                    params,
                    operations,
                    entangler,
                    layer_index,
                    rep,
                    max_reps,
                    control,
                    target,
                    spacing,
                    n_qubits,
                    is_corr,
                )
                tag(W_pairs, f"W{control,target}")
                rep_kron.append(W_pairs)

            layer_reps.append(kron(*rep_kron))

        # Combine all repetitions using `chain`
        current_layer.append(chain(*layer_reps))

    else:  # Original behavior for other layers
        for rep in range(max_reps):
            for control, target in zip(current_indices[::2], current_indices[1::2]):
                W_pairs = _create_single_W(
                    params,
                    operations,
                    entangler,
                    layer_index,
                    rep,
                    max_reps,
                    control,
                    target,
                    spacing,
                    n_qubits,
                    is_corr,
                )
                current_layer.append(W_pairs)

    # Update `next_indices` with the **targets** of the current layer
    next_indices = current_indices[1::2]
    return chain(*current_layer), next_indices



from __future__ import annotations

from typing import Any, Callable
from qadence.blocks import chain
from qadence.parameters import Parameter
from qadence.blocks.utils import add, tag
from qadence.operations import Z, RX, CZ
from qadence.circuit import QuantumCircuit
from qadence.types import BackendName, DiffMode
from qadence.blocks.abstract import AbstractBlock

from .qnn_model import QNN
from .qnn_constructors import _create_conv_layer, _create_feature_map_qcnn


class QCNN(QNN):
    def __init__(
        self,
        n_inputs: int,
        n_qubits: int,
        depth: list[int],
        operations: list[Any],
        entangler: Any = CZ,
        random_meas: bool = True,
        fm_basis: str = "Fourier",
        fm_gate: Any = RX,
        is_corr: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Creates a QCNN model.

        Args:
            n_inputs (int): Number of input features.
            n_qubits (int): Total number of qubits.
            depth (list[int]): List defining the depth (repetitions) of each layer.
            operations (list[Any]): List of quantum operations to apply
                in the gates (e.g., [RX, RZ]).
            entangler (Any): Entangling operation, such as CZ.
            random_meas (bool): If True, applies random weighted measurements.
            fm_basis (str): feature map basis.
            fm_gate (Any): gate employed in the fm, such as.
            **kwargs (Any): Additional keyword arguments for the parent QNN class.
        """
        self.n_inputs = n_inputs
        self.n_qubits = n_qubits
        self.depth = depth
        self.operations = operations
        self.entangler = entangler
        self.random_meas = random_meas
        self.fm_basis = fm_basis
        self.fm_gate = fm_gate
        self.is_corr = is_corr

        circuit = self.qcnn_circuit(
            self.n_inputs,
            self.n_qubits,
            self.depth,
            self.operations,
            self.entangler,
            self.fm_basis,
            self.fm_gate,
            self.is_corr,
        )

        obs = self.qcnn_deferred_obs(self.n_qubits, self.random_meas)

        super().__init__(
            circuit=circuit,
            observable=obs,
            backend=BackendName.PYQTORCH,
            diff_mode=DiffMode.AD,
            inputs=[f"\u03c6_{i}" for i in range(self.n_inputs)],
            **kwargs,
        )

    def qcnn_circuit(
        self,
        n_inputs: int,
        n_qubits: int,
        depth: list[int],
        operations: list[Any],
        entangler: AbstractBlock,
        fm_basis: str,
        fm_gate: AbstractBlock,
        is_corr: bool,
    ) -> QuantumCircuit:
        """Defines the QCNN circuit."""
        # Validate qubit count
        if n_qubits < 4:
            raise ValueError(
                f"Invalid number of qubits: {n_qubits}. " "At least 4 qubits are required."
            )
        if n_qubits % 2 != 0:
            raise ValueError(
                f"Invalid number of qubits: {n_qubits}. " "The number of qubits must be even."
            )

        # Validate that all values in `depth` are odd
        even_depths = [d for d in depth if d % 2 == 0]
        if even_depths:
            raise ValueError(
                f"Invalid depth values: '{even_depths[0]}'. " "All the conv layer 'r's must be odd."
            )

        # Feature map (FM)
        fm = _create_feature_map_qcnn(n_qubits, n_inputs, fm_basis, fm_gate)
        tag(fm, "FM")

        # Conv and Pool layer definition
        conv_layers = []
        params: dict[str, Parameter] = {}

        # Define layer all the 2-qubit patterns based on depth
        layer_patterns = [(2**layer_index, depth[layer_index]) for layer_index in range(len(depth))]

        # Initialize all qubits for the current layer
        current_indices = list(range(n_qubits))

        # Build the circuit layer by layer using the helper
        for layer_index, (_, reps) in enumerate(layer_patterns):
            if reps == 0:
                raise ValueError(f"Invalid layer {layer_index}: zero repetitions (reps = {reps}).")
            if len(current_indices) < 2:
                raise RuntimeError(
                    f"Layer {layer_index} requires at least 2 qubits, "
                    f"but found {len(current_indices)}."
                )

            layer_block, next_indices = _create_conv_layer(
                layer_index, reps, current_indices, params, operations, entangler, n_qubits, is_corr
            )
            tag(layer_block, f"C+P layer {layer_index}")
            conv_layers.append(layer_block)

            # Update `current_indices` for the next layer
            current_indices = next_indices

        # Combine all layers for the final ansatz
        ansatz = chain(*conv_layers)

        return QuantumCircuit(n_qubits, fm, ansatz)

    def qcnn_deferred_obs(
        self, n_qubits: int, random_meas: bool
    ) -> AbstractBlock | list[AbstractBlock]:
        """
        Defines the measurements to be performedthe traced out.

        and remaining qubits.
        """
        if random_meas:
            w1 = [Parameter(f"w{i}") for i in range(n_qubits)]
            obs = add(Z(i) * w for i, w in zip(range(n_qubits), w1))
        else:
            obs = add(Z(i) for i in range(n_qubits))

        return obs



from __future__ import annotations

from collections import Counter, OrderedDict
from logging import getLogger
from typing import Any, Callable

import sympy
import torch
from torch import Tensor, nn

from qadence.backend import BackendConfiguration, ConvertedObservable
from qadence.backends.api import config_factory
from qadence.blocks.abstract import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.mitigations import Mitigations
from qadence.model import QuantumModel
from qadence.noise import NoiseHandler
from qadence.register import Register
from qadence.types import BackendName, DiffMode, Endianness, InputDiffMode, ParamDictType
from qadence.utils import block_to_mathematical_expression
from .qnn_config import AnsatzConfig, FeatureMapConfig

logger = getLogger(__name__)


def _torch_derivative(
    ufa: Callable, x: torch.Tensor, derivative_indices: tuple[int, ...]
) -> torch.Tensor:
    y = ufa(x)
    for idx in derivative_indices:
        out = torch.autograd.grad(y, x, torch.ones_like(y), create_graph=True)[0]
        y = out[:, idx]
    return y.reshape(-1, 1)


def derivative(ufa: torch.nn.Module, x: Tensor, derivative_indices: tuple[int, ...]) -> Tensor:
    """Compute derivatives w.r.t.

    inputs of a UFA with a single output. The
    `derivative_indices` specify which derivative(s) are computed.  E.g.
    `derivative_indices=(1,2)` would compute the a second order derivative w.r.t
    to the indices `1` and `2` of the input tensor.

    Arguments:
        ufa: The model for which we want to compute the derivative.
        x (Tensor): (batch_size, input_size) input tensor.
        derivative_indices (tuple): Define which derivatives to compute.

    Examples:
    If we create a UFA with three inputs and denote the first, second, and third
    input with `x`, `y`, and `z` we can compute the following derivatives w.r.t
    to those inputs:
    ```py exec="on" source="material-block"
    import torch
    from qadence.ml_tools.models import derivative, QNN
    from qadence.ml_tools.config import FeatureMapConfig, AnsatzConfig
    from qadence.constructors.hamiltonians import ObservableConfig
    from qadence.operations import Z

    fm_config = FeatureMapConfig(num_features=3, inputs=["x", "y", "z"])
    ansatz_config = AnsatzConfig()
    obs_config = ObservableConfig(detuning=Z)

    f = QNN.from_configs(
        register=3, obs_config=obs_config, fm_config=fm_config, ansatz_config=ansatz_config,
    )
    inputs = torch.rand(5,3,requires_grad=True)

    # df_dx
    derivative(f, inputs, (0,))

    # d2f_dydz
    derivative(f, inputs, (1,2))

    # d3fdy2dx
    derivative(f, inputs, (1,1,0))
    ```
    """
    assert ufa.out_features == 1, "Can only call `derivative` on models with 1D output."
    return ufa._derivative(x, derivative_indices)


def format_to_dict_fn(
    inputs: list[sympy.Symbol | str] = [],
) -> Callable[[Tensor | ParamDictType], ParamDictType]:
    """Format an input tensor into the format required by the forward pass.

    The tensor is assumed to have dimensions: n_batches x in_features where in_features
    corresponds to the number of input features of the QNN
    """
    in_features = len(inputs)

    def tensor_to_dict(values: Tensor | ParamDictType) -> ParamDictType:
        if isinstance(values, Tensor):
            values = values.reshape(-1, 1) if len(values.size()) == 1 else values
            if not values.shape[1] == in_features:
                raise ValueError(
                    f"Model expects in_features={in_features} but got {values.shape[1]}."
                )
            values = {fparam.name: values[:, inputs.index(fparam)] for fparam in inputs}  # type: ignore[union-attr]
        return values

    return tensor_to_dict


class QNN(QuantumModel):
    """Quantum neural network model for n-dimensional inputs.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import QuantumCircuit, QNN, Z
    from qadence import hea, feature_map, hamiltonian_factory, kron

    # create the circuit
    n_qubits, depth = 2, 4
    fm = kron(
        feature_map(1, support=(0,), param="x"),
        feature_map(1, support=(1,), param="y")
    )
    ansatz = hea(n_qubits=n_qubits, depth=depth)
    circuit = QuantumCircuit(n_qubits, fm, ansatz)
    obs_base = hamiltonian_factory(n_qubits, detuning=Z)

    # the QNN will yield two outputs
    obs = [2.0 * obs_base, 4.0 * obs_base]

    # initialize and use the model
    qnn = QNN(circuit, obs, inputs=["x", "y"])
    y = qnn(torch.rand(3, 2))
    print(str(y)) # markdown-exec: hide
    ```
    """

    inputs: list[sympy.Basic | str]

    def __init__(
        self,
        circuit: QuantumCircuit,
        observable: list[AbstractBlock] | AbstractBlock,
        backend: BackendName = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        measurement: Measurements | None = None,
        noise: NoiseHandler | None = None,
        configuration: BackendConfiguration | dict | None = None,
        inputs: list[sympy.Basic | str] | None = None,
        input_diff_mode: InputDiffMode | str = InputDiffMode.AD,
    ):
        """Initialize the QNN.

        The number of inputs is determined by the feature parameters in the input
        quantum circuit while the number of outputs is determined by how many
        observables are provided as input

        Args:
            circuit: The quantum circuit to use for the QNN.
            observable: The observable.
            backend: The chosen quantum backend.
            diff_mode: The differentiation engine to use. Choices 'gpsr' or 'ad'.
            measurement: optional measurement protocol. If None,
                use exact expectation value with a statevector simulator
            noise: A noise model to use.
            configuration: optional configuration for the backend
            inputs: List that indicates the order of variables of the tensors that are passed
                to the model. Given input tensors `xs = torch.rand(batch_size, input_size:=2)` a QNN
                with `inputs=["t", "x"]` will assign `t, x = xs[:,0], xs[:,1]`.
            input_diff_mode: The differentiation mode for the input tensor.
        """
        super().__init__(
            circuit,
            observable=observable,
            backend=backend,
            diff_mode=diff_mode,
            measurement=measurement,
            configuration=configuration,
            noise=noise,
        )
        if self._observable is None:
            raise ValueError("You need to provide at least one observable in the QNN constructor")
        if (inputs is not None) and (len(self.inputs) == len(inputs)):
            self.inputs = [sympy.symbols(x) if isinstance(x, str) else x for x in inputs]  # type: ignore[union-attr]
        elif (inputs is None) and len(self.inputs) <= 1:
            self.inputs = [sympy.symbols(x) if isinstance(x, str) else x for x in self.inputs]  # type: ignore[union-attr]
        else:
            raise ValueError(
                """
                Your QNN has more than one input. Please provide a list of inputs in the order of
                your tensor domain. For example, if you want to pass
                `xs = torch.rand(batch_size, input_size:=3)` to you QNN, where
                ```
                t = x[:,0]
                x = x[:,1]
                y = x[:,2]
                ```
                you have to specify
                ```
                QNN(circuit, observable, inputs=["t", "x", "y"])
                ```
                You can also pass a list of sympy symbols.
            """
            )
        self.format_to_dict = format_to_dict_fn(self.inputs)  # type: ignore[arg-type]
        self.input_diff_mode = InputDiffMode(input_diff_mode)
        if self.input_diff_mode == InputDiffMode.FD:
            from qadence.backends.utils import finitediff

            self.__derivative = finitediff
        elif self.input_diff_mode == InputDiffMode.AD:
            self.__derivative = _torch_derivative  # type: ignore[assignment]
        else:
            raise ValueError(f"Unkown forward diff mode: {self.input_diff_mode}")

        self._model_configs: dict = dict()

    @classmethod
    def from_configs(
        cls,
        register: int | Register,
        obs_config: Any,
        fm_config: Any = FeatureMapConfig(),
        ansatz_config: Any = AnsatzConfig(),
        backend: BackendName = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        measurement: Measurements | None = None,
        noise: NoiseHandler | None = None,
        configuration: BackendConfiguration | dict | None = None,
        input_diff_mode: InputDiffMode | str = InputDiffMode.AD,
    ) -> QNN:
        """Create a QNN from a set of configurations.

        Args:
            register (int | Register): The number of qubits or a register object.
            obs_config (list[ObservableConfig] | ObservableConfig): The configuration(s)
                for the observable(s).
            fm_config (FeatureMapConfig): The configuration for the feature map.
                Defaults to no feature encoding block.
            ansatz_config (AnsatzConfig): The configuration for the ansatz.
                Defaults to a single layer of hardware efficient ansatz.
            backend (BackendName): The chosen quantum backend.
            diff_mode (DiffMode): The differentiation engine to use. Choices are
                'gpsr' or 'ad'.
            measurement (Measurements): Optional measurement protocol. If None,
                use exact expectation value with a statevector simulator.
            noise (Noise): A noise model to use.
            configuration (BackendConfiguration | dict): Optional backend configuration.
            input_diff_mode (InputDiffMode): The differentiation mode for the input tensor.

        Returns:
            A QNN object.

        Raises:
            ValueError: If the observable configuration is not provided.

        Example:
        ```python exec="on" source="material-block" result="json"
        import torch
        from qadence.ml_tools.config import AnsatzConfig, FeatureMapConfig
        from qadence.ml_tools import QNN
        from qadence.constructors import ObservableConfig
        from qadence.operations import Z
        from qadence.types import (
            AnsatzType, BackendName, BasisSet, ReuploadScaling, Strategy
        )

        register = 4
        obs_config = ObservableConfig(
            detuning=Z,
            scale=5.0,
            shift=0.0,
            trainable_transform=None,
        )
        fm_config = FeatureMapConfig(
            num_features=2,
            inputs=["x", "y"],
            basis_set=BasisSet.FOURIER,
            reupload_scaling=ReuploadScaling.CONSTANT,
            feature_range={
                "x": (-1.0, 1.0),
                "y": (0.0, 1.0),
            },
        )
        ansatz_config = AnsatzConfig(
            depth=2,
            ansatz_type=AnsatzType.HEA,
            ansatz_strategy=Strategy.DIGITAL,
        )

        qnn = QNN.from_configs(
            register, obs_config, fm_config, ansatz_config, backend=BackendName.PYQTORCH
        )

        x = torch.rand(2, 2)
        y = qnn(x)
        print(str(y)) # markdown-exec: hide
        ```
        """
        from .qnn_constructors import build_qnn_from_configs

        qnn = build_qnn_from_configs(
            register=register,
            observable_config=obs_config,
            fm_config=fm_config,
            ansatz_config=ansatz_config,
            backend=backend,
            diff_mode=diff_mode,
            measurement=measurement,
            noise=noise,
            configuration=configuration,
            input_diff_mode=input_diff_mode,
        )
        qnn._model_configs = {
            "register": register,
            "observable_config": obs_config,
            "fm_config": fm_config,
            "ansatz_config": ansatz_config,
        }
        return qnn

    def __str__(self) -> str | Any:
        """Return a string representation of a QNN.

        When creating a QNN from a set of configurations,
        we print the configurations used. Otherwise, we use the default printing.

        Returns:
            str | Any: A string representation of a QNN.

        Example:
        ```python exec="on" source="material-block" result="json"
        from qadence import QNN
        from qadence.constructors.hamiltonians import Interaction
        from qadence.ml_tools.config import AnsatzConfig, FeatureMapConfig
        from qadence.ml_tools.constructors import (
            ObservableConfig,
        )
        from qadence.operations import Z
        from qadence.types import BackendName

        backend = BackendName.PYQTORCH
        fm_config = FeatureMapConfig(num_features=1)
        ansatz_config = AnsatzConfig()
        observable_config = ObservableConfig(detuning=Z, interaction=Interaction.ZZ, scale=2)

        qnn = QNN.from_configs(
            register=2,
            obs_config=observable_config,
            fm_config=fm_config,
            ansatz_config=ansatz_config,
            backend=backend,
        )
        print(qnn) # markdown-exec: hide
        ```
        """
        if bool(self._model_configs):
            configs_str = "\n".join(
                (
                    k + " = " + str(self._model_configs[k])
                    for k in sorted(self._model_configs.keys())
                    if k != "observable_config"
                )
            )
            observable_str = ""
            if self._observable:
                observable_str = f"observable_config = {self.observables_to_expression()}"

            return f"{type(self).__name__}(\n{configs_str}\n{observable_str}\n)"

        return super().__str__()

    def forward(
        self,
        values: dict[str, Tensor] | Tensor = None,
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: NoiseHandler | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """Forward pass of the model.

        This returns the (differentiable) expectation value of the given observable
        operator defined in the constructor. Differently from the base QuantumModel
        class, the QNN accepts also a tensor as input for the forward pass. The
        tensor is expected to have shape: `n_batches x in_features` where `n_batches`
        is the number of data points and `in_features` is the dimensionality of the problem

        The output of the forward pass is the expectation value of the input
        observable(s). If a single observable is given, the output shape is
        `n_batches` while if multiple observables are given the output shape
        is instead `n_batches x n_observables`

        Args:
            values: the values of the feature parameters
            state: Initial state.
            measurement: optional measurement protocol. If None,
                use exact expectation value with a statevector simulator
            noise: A noise model to use.
            endianness: Endianness of the resulting bit strings.

        Returns:
            Tensor: a tensor with the expectation value of the observables passed
                in the constructor of the model
        """
        return self.expectation(
            values, state=state, measurement=measurement, noise=noise, endianness=endianness
        )

    def run(
        self,
        values: Tensor | dict[str, Tensor] = None,
        state: Tensor | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        return super().run(
            values=self.format_to_dict(values),
            state=state,
            endianness=endianness,
        )

    def sample(
        self,
        values: Tensor | dict[str, Tensor] = {},
        n_shots: int = 1000,
        state: Tensor | None = None,
        noise: NoiseHandler | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> list[Counter]:
        return super().sample(  # type: ignore[no-any-return]
            values=self.format_to_dict(values),
            n_shots=n_shots,
            state=state,
            noise=noise,
            mitigation=mitigation,
            endianness=endianness,
        )

    def expectation(
        self,
        values: Tensor | dict[str, Tensor] = {},
        observable: list[ConvertedObservable] | ConvertedObservable | None = None,
        state: Tensor | None = None,
        measurement: Measurements | None = None,
        noise: NoiseHandler | None = None,
        mitigation: Mitigations | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        if values is None:
            values = {}
        if measurement is None:
            measurement = self._measurement
        if noise is None:
            noise = self._noise
        return super().expectation(
            values=self.format_to_dict(values),
            state=state,
            measurement=measurement,
            endianness=endianness,
            noise=noise,
        )

    def _derivative(self, x: Tensor, derivative_indices: tuple[int, ...]) -> Tensor:
        return self.__derivative(self, x, derivative_indices)

    def _to_dict(self, save_params: bool = False) -> dict:
        d = dict()
        try:
            d = super()._to_dict(save_params)
            d[self.__class__.__name__]["inputs"] = [str(i) for i in self.inputs]
            logger.debug(f"{self.__class__.__name__} serialized to {d}.")
        except Exception as e:
            logger.warning(f"Unable to serialize {self.__class__.__name__} due to {e}.")
        return d

    @classmethod
    def _from_dict(cls, d: dict, as_torch: bool = False) -> QNN:
        from qadence.serialization import deserialize

        qnn: QNN
        try:
            qm_dict = d[cls.__name__]
            qnn = cls(
                circuit=QuantumCircuit._from_dict(qm_dict["circuit"]),
                observable=[deserialize(q_obs) for q_obs in qm_dict["observable"]],  # type: ignore[misc]
                backend=qm_dict["backend"],
                diff_mode=qm_dict["diff_mode"],
                measurement=Measurements._from_dict(qm_dict["measurement"]),
                noise=NoiseHandler._from_dict(qm_dict["noise"]),
                configuration=config_factory(qm_dict["backend"], qm_dict["backend_configuration"]),
                inputs=qm_dict["inputs"],
            )

            if as_torch:
                conv_pd = nn.ParameterDict()
                param_dict = d["param_dict"]
                for n, param in param_dict.items():
                    conv_pd[n] = nn.Parameter(param)
                qnn._params = conv_pd
            logger.debug(f"Initialized {cls.__name__} from {d}.")

        except Exception as e:
            logger.warning(f"Unable to deserialize object {d} to {cls.__name__} due to {e}.")

        return qnn



from __future__ import annotations

from .qfi import get_quantum_fisher, get_quantum_fisher_spsa
from .qng import QuantumNaturalGradient

# Modules to be automatically added to the qadence namespace
__all__ = [
    "QuantumNaturalGradient",
    "get_quantum_fisher",
    "get_quantum_fisher_spsa",
]



from __future__ import annotations

import torch
from qadence import Overlap
from torch import Tensor


def _create_random_direction(size: int) -> Tensor:
    """
    Creates a torch Tensor with `size` elements randomly drawn from {-1,+1}.

    Args:
        size (int): Size of the vector
    """
    x = torch.bernoulli(0.5 * torch.ones(size)).reshape(size, 1)
    x[x == 0] = -1
    return x


def _shifted_overlap(
    model: Overlap,
    shift: Tensor,
    fm_dict: dict[str, Tensor],
    vparams_dict: dict[str, Tensor],
) -> Tensor:
    """
    Forward method of the Overlap model with shifted values of the ket variational parameters.

    Args:
        model (Overlap): Overlap model
        shift (float): Quantity of the shift
        fm_dict (dict[str, Tensor]): Feature map dictionary
        vparams_dict (dict[str, Tensor]): Variational parameter dictionary
    """
    shifted_vparams_dict = {k: (v + s) for (k, v), s in zip(vparams_dict.items(), shift)}

    ovrlp_shifted = model(
        bra_param_values=fm_dict | vparams_dict,
        ket_param_values=fm_dict | shifted_vparams_dict,
    )
    return ovrlp_shifted


def spsa_gradient_step(
    model: Overlap,
    epsilon: float,
    fm_dict: dict[str, Tensor],
    vparams_dict: dict[str, Tensor] = dict(),
) -> Tensor:
    """Single step of the first order SPSA gradient.

    Calculates a single step of the SPSA algorithm to calculate
    the first order gradient of the given Overlap model.

    Args:
        model (Overlap): Overlap model
        epsilon (float): Finite step size
        fm_dict (dict[str, Tensor]): Feature map dictionary
        vparams_dict (dict[str, Tensor]): Variational parameters dictionary
    """
    if not vparams_dict:
        vparams_dict = {k: v for (k, v) in model._params.items() if v.requires_grad}

    # Create random direction
    random_direction = _create_random_direction(size=model.num_vparams)

    # Shift ket variational parameters
    shift = epsilon * random_direction

    # Overlaps with the shifted parameters
    ovrlp_shifted_plus = _shifted_overlap(model, shift, fm_dict, vparams_dict)
    ovrlp_shifted_minus = _shifted_overlap(model, -shift, fm_dict, vparams_dict)

    return random_direction * (ovrlp_shifted_plus - ovrlp_shifted_minus) / (2 * epsilon)


def spsa_2gradient_step(
    model: Overlap,
    epsilon: float,
    fm_dict: dict[str, Tensor],
    vparams_dict: dict[str, Tensor] = dict(),
) -> Tensor:
    """Single step of the second order SPSA gradient.

    Calculates a single step of the SPSA algorithm to calculate
    the second order gradient of the given Overlap model.

    TODO: implement recursively using the first order function

    Args:
        model (Overlap): Overlap model
        epsilon (float): Finite step size
        fm_dict (dict[str, Tensor]): Feature map dictionary
        vparams_dict (dict[str, Tensor]): Variational parameters dictionary
    """
    if not vparams_dict:
        vparams_dict = {k: v for (k, v) in model._params.items() if v.requires_grad}

    # Create random directions
    rand_dir1 = _create_random_direction(size=model.num_vparams)
    rand_dir2 = _create_random_direction(size=model.num_vparams)

    # Overlaps with the shifted parameters
    shift_p1 = epsilon * rand_dir1
    ovrlp_shifted_p1 = _shifted_overlap(model, shift_p1, fm_dict, vparams_dict)
    shift_p1p2 = epsilon * (rand_dir1 + rand_dir2)
    ovrlp_shifted_p1p2 = _shifted_overlap(model, shift_p1p2, fm_dict, vparams_dict)
    shift_m1 = -epsilon * rand_dir1
    ovrlp_shifted_m1 = _shifted_overlap(model, shift_m1, fm_dict, vparams_dict)
    shift_m1p2 = epsilon * (-rand_dir1 + rand_dir2)
    ovrlp_shifted_m1p2 = _shifted_overlap(model, shift_m1p2, fm_dict, vparams_dict)

    # Prefactor
    delta_F = ovrlp_shifted_p1p2 - ovrlp_shifted_p1 - ovrlp_shifted_m1p2 + ovrlp_shifted_m1

    # Hessian
    dir_product = torch.matmul(rand_dir1, rand_dir2.transpose(0, 1)) + torch.matmul(
        rand_dir2, rand_dir1.transpose(0, 1)
    )
    hess = (4 * (epsilon**2)) ** (-1) * delta_F * dir_product

    return hess



from __future__ import annotations

import torch
from qadence import Overlap, QuantumCircuit
from qadence.blocks import parameters, primitive_blocks
from qadence.types import BackendName, DiffMode, OverlapMethod
from torch import Tensor

from qadence_model.optimizers.spsa import spsa_2gradient_step
from qadence_model.optimizers.utils import hessian


def _positive_semidefinite_sqrt(A: Tensor) -> Tensor:
    """Computes the square root of a real positive semi-definite matrix.

    Code taken from https://github.com/pytorch/pytorch/issues/25481#issuecomment-1032789228
    """
    L, Q = torch.linalg.eigh(A)
    zero = torch.zeros((), device=L.device, dtype=L.dtype)
    threshold = L.max(-1).values * L.size(-1) * torch.finfo(L.dtype).eps
    L = L.where(L > threshold.unsqueeze(-1), zero)  # zero out small components
    return (Q * L.sqrt().unsqueeze(-2)) @ Q.mH


def _set_circuit_vparams(circuit: QuantumCircuit, vparams_dict: dict[str, Tensor]) -> None:
    """Sets the variational parameter values of the circuit."""
    blocks = primitive_blocks(circuit.block)
    for block in blocks:
        params = parameters(block)
        for p in params:
            if p.trainable:
                p.value = vparams_dict[p.name]


def _get_fm_dict(circuit: QuantumCircuit) -> dict[str, Tensor]:
    """Returns a dictionary holding the FM parameters of the circuit."""
    fm_dict = {}
    blocks = primitive_blocks(circuit.block)
    for block in blocks:
        params = parameters(block)
        for p in params:
            if not p.trainable:
                fm_dict[p.name] = torch.tensor([p.value])
    return fm_dict


def get_quantum_fisher(
    circuit: QuantumCircuit,
    vparams_dict: dict[str, Tensor] = dict(),
    fm_dict: dict[str, Tensor] = dict(),
    backend: BackendName = BackendName.PYQTORCH,
    overlap_method: OverlapMethod = OverlapMethod.EXACT,
    diff_mode: DiffMode = DiffMode.AD,
) -> Tensor:
    """Returns the exact Quantum Fisher Information (QFI) matrix.

    Args:
        circuit (QuantumCircuit): The quantum circuit.
        vparams_dict (dict[str, Tensor]):
            Dictionary holding the values of the variational parameters of the circuit.
        fm_dict (dict[str, Tensor]):
            Dictionary holding the values of the Feature Map parameters of the circuit.
        overlap_method (OverlapMethod, optional): Defaults to OverlapMethod.EXACT.
        diff_mode (DiffMode, optional): Defaults to DiffMode.ad.

    Returns:
        Tensor:
            The exact QFI matrix of the circuit.
    """
    # The FM dictionary is needed for the forward method in Overlap()
    if not fm_dict:
        fm_dict = _get_fm_dict(circuit)

    if vparams_dict:
        _set_circuit_vparams(circuit, vparams_dict)

    overlap_model = Overlap(
        circuit,
        circuit,
        backend=backend,
        diff_mode=diff_mode,
        method=overlap_method,
    )
    overlap = overlap_model(bra_param_values=fm_dict, ket_param_values=fm_dict)

    # Retrieve variational parameters of the overlap model
    # Importantly, the vparams of the overlap model are the vparams of the bra tensor,
    # Which means if we differentiate wrt vparams we are differentiating only wrt the
    # parameters in the bra and not in the ket
    vparams = [v for v in overlap_model._params.values() if v.requires_grad]
    return -2 * hessian(overlap, vparams)


def get_quantum_fisher_spsa(
    circuit: QuantumCircuit,
    iteration: int,
    vparams_dict: dict[str, Tensor] = dict(),
    fm_dict: dict[str, Tensor] = dict(),
    previous_qfi_estimator: Tensor | None = None,
    epsilon: float = 10e-3,
    beta: float = 10e-2,
    backend: BackendName = BackendName.PYQTORCH,
    overlap_method: OverlapMethod = OverlapMethod.EXACT,
    diff_mode: DiffMode = DiffMode.AD,
) -> tuple[Tensor, Tensor]:
    """Returns the a SPSA-approximation of the Quantum Fisher Information (QFI) matrix.

    Args:
        circuit (QuantumCircuit): The quantum circuit.
        iteration (int): Current iteration in the SPSA iterative loop.
        vparams_values (dict[str, Tensor]):
            Dictionary holding the values of the variational parameters of the circuit.
        fm_dict (dict[str, Tensor]):
            Dictionary holding the values of the Feature Map parameters of the circuit.
        overla p_method (OverlapMethod, optional): Defaults to OverlapMethod.EXACT.
        diff_mode (DiffMode, optional): Defaults to DiffMode.ad.

    Returns:
        tuple[Tensor, Tensor]:
            Tuple containing the QFI matrix and its positive semi-definite estimator.
    """
    # Retrieve feature parameters if they are not given as inputs
    # The FM dictionary is needed for the forward run of the Overlap() model
    if not fm_dict:
        fm_dict = _get_fm_dict(circuit)

    ovrlp_model = Overlap(
        circuit,
        circuit,
        backend=backend,
        diff_mode=diff_mode,
        method=overlap_method,
    )

    # Calculate the QFI matrix
    qfi_mat = -2 * spsa_2gradient_step(ovrlp_model, epsilon, fm_dict, vparams_dict)

    # Calculate the QFI estimator from the old estimator of qfi_mat
    if iteration == 0:
        qfi_mat_estimator = qfi_mat
    else:
        a_k = 1 / (1 + iteration)
        qfi_mat_estimator = a_k * (iteration * previous_qfi_estimator + qfi_mat)  # type: ignore

    # Get the positive-semidefinite version of the matrix for the update rule in QNG
    qfi_mat_positive_sd = _positive_semidefinite_sqrt(qfi_mat_estimator @ qfi_mat_estimator)
    qfi_mat_positive_sd = qfi_mat_positive_sd + beta * torch.eye(ovrlp_model.num_vparams)
    qfi_mat_positive_sd = qfi_mat_positive_sd / (1 + beta)  # regularization

    return qfi_mat_estimator, qfi_mat_positive_sd



from __future__ import annotations

from typing import Any

import torch
from qadence.blocks import AbstractBlock, add, chain, kron, tag
from qadence.constructors import hamiltonian_factory
from qadence.constructors.daqc import daqc_transform
from qadence.operations import CPHASE, SWAP, H, HamEvo, I, Z
from qadence.types import PI, Interaction, Strategy
from typing import Optional


def qft(
    n_qubits: int,
    support: Optional[tuple[int, ...]] = None,
    inverse: bool = False,
    reverse_in: bool = False,
    swaps_out: bool = False,
    strategy: Strategy = Strategy.DIGITAL,
    gen_build: AbstractBlock | None = None,
) -> AbstractBlock:
    """
    The Quantum Fourier Transform.

    Depending on the application, user should be careful with qubit ordering
    in the input and output. This can be controlled with reverse_in and swaps_out
    arguments.

    Args:
        n_qubits: number of qubits in the QFT
        support: qubit support to use
        inverse: True performs the inverse QFT
        reverse_in: Reverses the input qubits to account for endianness
        swaps_out: Performs swaps on the output qubits to match the "textbook" QFT.
        strategy: Strategy.Digital or Strategy.sDAQC
        gen_build: building block Ising Hamiltonian for the DAQC transform.
            Defaults to constant all-to-all Ising.

    Examples:
        ```python exec="on" source="material-block" result="json"
        from qadence_gradient import qft

        n_qubits = 3

        qft_circuit = qft(n_qubits, strategy = "sDAQC")
        ```
    """

    if support is None:
        support = tuple(range(n_qubits))

    assert len(support) <= n_qubits, "Wrong qubit support supplied"

    if reverse_in:
        support = support[::-1]

    qft_layer_dict = {
        Strategy.DIGITAL: _qft_layer_digital,
        Strategy.SDAQC: _qft_layer_sDAQC,
        Strategy.BDAQC: _qft_layer_bDAQC,
        Strategy.ANALOG: _qft_layer_analog,
    }

    try:
        layer_func = qft_layer_dict[strategy]
    except KeyError:
        raise KeyError(f"Strategy {strategy} not recognized.")

    qft_layers = reversed(range(n_qubits)) if inverse else range(n_qubits)

    qft_circ = chain(
        layer_func(
            n_qubits=n_qubits,
            support=support,
            layer=layer,
            inverse=inverse,
            gen_build=gen_build,
        )  # type: ignore
        for layer in qft_layers
    )

    if swaps_out:
        swap_ops = [SWAP(support[i], support[n_qubits - i - 1]) for i in range(n_qubits // 2)]
        qft_circ = chain(*swap_ops, qft_circ) if inverse else chain(qft_circ, *swap_ops)

    return tag(qft_circ, tag="iQFT") if inverse else tag(qft_circ, tag="QFT")


########################
# STANDARD DIGITAL QFT #
########################


def _qft_layer_digital(
    n_qubits: int,
    support: tuple[int, ...],
    layer: int,
    inverse: bool,
    gen_build: AbstractBlock | None = None,
) -> AbstractBlock:
    """
    Apply the Hadamard gate followed by CPHASE gates.

    This corresponds to one layer of the QFT.
    """
    qubit_range_layer = (
        reversed(range(layer + 1, n_qubits)) if inverse else range(layer + 1, n_qubits)
    )
    rots = []
    for j in qubit_range_layer:
        angle = torch.tensor(
            ((-1) ** inverse) * 2 * PI / (2 ** (j - layer + 1)), dtype=torch.cdouble
        )
        rots.append(CPHASE(support[j], support[layer], angle))  # type: ignore
    if inverse:
        return chain(*rots, H(support[layer]))  # type: ignore
    return chain(H(support[layer]), *rots)  # type: ignore


########################################
# DIGITAL-ANALOG QFT (with sDAQC)      #
# [1] https://arxiv.org/abs/1906.07635 #
########################################


def _theta(k: int) -> float:
    """Equation (16) from [1]."""
    return float(PI / (2 ** (k + 1)))


def _alpha(c: int, m: int, k: int) -> float:
    """Equation (16) from [1]."""
    if c == m:
        return float(PI / (2 ** (k - m + 2)))
    else:
        return 0.0


def _sqg_gen(n_qubits: int, support: tuple[int, ...], m: int, inverse: bool) -> list[AbstractBlock]:
    """Equation (13) from [1].

    Creates the generator corresponding to single-qubit rotations coming
    out of the CPHASE decomposition. The paper also includes the generator
    for the Hadamard of each layer here, but we left it explicit at
    the start of each layer.
    """
    k_sqg_list = reversed(range(2, n_qubits - m + 2)) if inverse else range(2, n_qubits - m + 2)

    sqg_gen_list = []
    for k in k_sqg_list:
        sqg_gen = (
            kron(I(support[j]) for j in range(n_qubits)) - Z(support[k + m - 2]) - Z(support[m - 1])
        )
        sqg_gen_list.append(_theta(k) * sqg_gen)

    return sqg_gen_list


def _tqg_gen(n_qubits: int, support: tuple[int, ...], m: int, inverse: bool) -> list[AbstractBlock]:
    """Equation (14) from [1].

    Creates the generator corresponding to the two-qubit ZZ
    interactions coming out of the CPHASE decomposition.
    """
    k_tqg_list = reversed(range(2, n_qubits + 1)) if inverse else range(2, n_qubits + 1)

    tqg_gen_list = []
    for k in k_tqg_list:
        for c in range(1, k):
            tqg_gen = kron(Z(support[c - 1]), Z(support[k - 1]))
            tqg_gen_list.append(_alpha(c, m, k) * tqg_gen)

    return tqg_gen_list


def _qft_layer_sDAQC(
    n_qubits: int,
    support: tuple[int, ...],
    layer: int,
    inverse: bool,
    gen_build: AbstractBlock | None,
) -> AbstractBlock:
    """
    QFT Layer using the sDAQC technique.

    Following the paper:

    -- [1] https://arxiv.org/abs/1906.07635

    4 - qubit edge case is not implemented.

    Note: the paper follows an index convention of running from 1 to N. A few functions
    here also use that convention to be consistent with the paper. However, for qadence
    related things the indices are converted to [0, N-1].
    """

    # TODO: Properly check and include support for changing qubit support
    allowed_support = tuple(range(n_qubits))
    if support != allowed_support and support != allowed_support[::-1]:
        raise NotImplementedError("Changing support for DigitalAnalog QFT not yet supported.")

    if gen_build is None:
        gen_build = hamiltonian_factory(n_qubits, interaction=Interaction.NN)

    m = layer + 1  # Paper index convention

    # Generator for the single-qubit rotations contributing to the CPHASE gate
    sqg_gen_list = _sqg_gen(n_qubits=n_qubits, support=support, m=m, inverse=inverse)

    # Ising model representing the CPHASE gates two-qubit interactions
    tqg_gen_list = _tqg_gen(n_qubits=n_qubits, support=support, m=m, inverse=inverse)

    if len(sqg_gen_list) > 0:
        # Single-qubit rotations (leaving the Hadamard explicit)
        sq_gate = chain(H(support[m - 1]), HamEvo(add(*sqg_gen_list), -1.0))

        # Two-qubit interaction in the CPHASE converted with sDAQC
        gen_cphases = add(*tqg_gen_list)
        transformed_daqc_circuit = daqc_transform(
            n_qubits=gen_build.n_qubits,
            gen_target=gen_cphases,
            t_f=-1.0,
            gen_build=gen_build,
        )

        layer_circ = chain(
            sq_gate,
            transformed_daqc_circuit,
        )
        if inverse:
            return layer_circ.dagger()
        return layer_circ  # type: ignore
    else:
        return chain(H(support[m - 1]))  # type: ignore


########################################
# DIGITAL-ANALOG QFT (with bDAQC)      #
# [1] https://arxiv.org/abs/1906.07635 #
########################################


def _qft_layer_bDAQC(*args: Any, **kwargs: Any) -> Any:
    raise NotImplementedError


############
## ANALOG ##
############


def _qft_layer_analog(*args: Any, **kwargs: Any) -> Any:
    raise NotImplementedError



from __future__ import annotations

from typing import Type, Union

import qadence as qd
import sympy
from qadence.blocks import AddBlock, ChainBlock, add, chain
from qadence.constructors import hamiltonian_factory
from qadence.operations import N, X, Y, Z
from qadence.parameters import Parameter, VariationalParameter
from typing import Optional


TPauliOp = Union[Type[X], Type[Y], Type[Z], Type[N]]


def _amplitude_map(
    n_qubits: int,
    pauli_op: TPauliOp,
    weights: list[Parameter] | list[float] | None = None,
) -> AddBlock:
    """Create an generator equivalent to a laser amplitude mapping on the device.

    Basically, given a certain quantum operation `pauli_op`, this routine constructs
    the following generator:

        H = sum_i^N w_i OP(i)

    where the weights are variational parameters

    Args:
        n_qubits: number of qubits
        pauli_op: type of Pauli operation to use when creating the generator
        weights: list of variational paramters with the weights

    Returns:
        A block with the Hamiltonian generator
    """
    if weights is None:
        return add(pauli_op(j) for j in range(n_qubits))
    else:
        assert len(weights) <= n_qubits, "Wrong weights supplied"
        return add(w * pauli_op(j) for j, w in enumerate(weights))  # type:ignore [operator]


def rydberg_hea_layer(
    register: qd.Register,
    tevo_drive: Parameter | float,
    tevo_det: Parameter | float,
    tevo_wait: Parameter | float,
    phase: Parameter | float | None = None,
    detunings: list[Parameter] | list[float] | None = None,
    drives: list[Parameter] | list[float] | None = None,
    drive_scaling: float = 1.0,
) -> ChainBlock:
    """A single layer of the Rydberg hardware efficient ansatz.

    Args:
        register: the input register with atomic coordinates needed to build the interaction.
        tevo_drive: a variational parameter for the duration of the drive term of
            the Hamiltonian generator, including optional semi-local addressing
        tevo_det: a variational parameter for the duration of the detuning term of the
            Hamiltonian generator, including optional semi-local addressing
        tevo_wait: a variational parameter for the duration of the waiting
            time with interaction only
        phase: a variational parameter representing the global phase. If None, the
            global phase is set to 0 which results in a drive term in sigma^x only. Otherwise
            both sigma^x and sigma^y terms will be present
        detunings: a list of parameters with the weights of the locally addressed
            detuning terms. These are variational parameters which are tuned by the optimizer
        drives: a list of parameters with the weights of the locally addressed
            drive terms. These are variational parameters which are tuned by the optimizer
        drive_scaling: a scaling term to be added to the drive Hamiltonian generator

    Returns:
        A block with a single layer of Rydberg HEA
    """
    n_qubits = register.n_qubits

    drive_x = _amplitude_map(n_qubits, qd.X, weights=drives)
    drive_y = _amplitude_map(n_qubits, qd.Y, weights=drives)
    detuning = _amplitude_map(n_qubits, qd.N, weights=detunings)
    interaction = hamiltonian_factory(register, qd.Interaction.NN)

    # drive and interaction are not commuting thus they need to be
    # added directly into the final Hamiltonian generator
    if phase is not None:
        generator = (
            drive_scaling * sympy.cos(phase) * drive_x
            - drive_scaling * sympy.sin(phase) * drive_y
            + interaction
        )
    else:
        generator = drive_scaling * drive_x + interaction

    return chain(
        qd.HamEvo(generator, tevo_drive),
        # detuning and interaction are commuting, so they
        # can be ordered arbitrarily and treated separately
        qd.HamEvo(interaction, tevo_wait),
        qd.HamEvo(detuning, tevo_det),
    )


def rydberg_hea(
    register: qd.Register,
    n_layers: int = 1,
    addressable_detuning: bool = True,
    addressable_drive: bool = False,
    tunable_phase: bool = False,
    additional_prefix: Optional[str] = None,
) -> qd.blocks.ChainBlock:
    """Hardware efficient ansatz for neutral atom (Rydberg) platforms.

    This constructor implements a variational ansatz which is very close to
    what is implementable on 2nd generation PASQAL quantum devices. In particular,
    it implements evolution over a specific Hamiltonian which can be realized on
    the device. This Hamiltonian contains:

    * an interaction term given by the standard NN interaction and determined starting
        from the positions in the input register: Hᵢₙₜ = ∑ᵢⱼ C₆/rᵢⱼ⁶ nᵢnⱼ

    * a detuning term which corresponding to a n_i = (1+sigma_i^z)/2 applied to
        all the qubits. If the `addressable_detuning` flag is set to True, the routine
        effectively a local n_i = (1+sigma_i^z)/2 term in the
        evolved Hamiltonian with a different coefficient for each atom. These
        coefficients determine a local addressing pattern for the detuning on a subset
        of the qubits. In this routine, the coefficients are variational parameters
        and they will therefore be optimized at each optimizer step

    * a drive term which corresponding to a sigma^x evolution operation applied to
        all the qubits. If the `addressable_drive` flag is set to True, the routine
        effectively a local sigma_i^x term in the evolved Hamiltonian with a different
        coefficient for each atom. These coefficients determine a local addressing pattern
        for the drive on a subset of the qubits. In this routine, the coefficients are
        variational parameters and they will therefore be optimized at each optimizer step

    * if the `tunable_phase` flag is set to True, the drive term is modified in the following
        way: drive = cos(phi) * sigma^x - sin(phi) * sigma^y
        The addressable pattern above is maintained and the phase is considered just as an
        additional variational parameter which is optimized with the rest

    Notice that, on real devices, the coefficients assigned to each qubit in both the detuning
    and drive patterns should be non-negative and they should always sum to 1. This is not the
    case for the implementation in this routine since the coefficients (weights) do not have any
    constraint. Therefore, this HEA is not completely realizable on neutral atom devices.

    Args:
        register: the input atomic register with Cartesian coordinates.
        n_layers: number layers in the HEA, each layer includes a drive, detuning and
            pure interaction pulses whose is a variational parameter
        addressable_detuning: whether to turn on the trainable semi-local addressing pattern
            on the detuning (n_i terms in the Hamiltonian)
        addressable_drive: whether to turn on the trainable semi-local addressing pattern
            on the drive (sigma_i^x terms in the Hamiltonian)
        tunable_phase: whether to have a tunable phase to get both sigma^x and sigma^y rotations
            in the drive term. If False, only a sigma^x term will be included in the drive part
            of the Hamiltonian generator
        additional_prefix: an additional prefix to attach to the parameter names

    Returns:
        The Rydberg HEA block
    """
    n_qubits = register.n_qubits
    prefix = "" if additional_prefix is None else "_" + additional_prefix

    detunings = None
    # add a detuning pattern locally addressing the atoms
    if addressable_detuning:
        detunings = [qd.VariationalParameter(f"detmap_{j}") for j in range(n_qubits)]

    drives = None
    # add a drive pattern locally addressing the atoms
    if addressable_drive:
        drives = [qd.VariationalParameter(f"drivemap_{j}") for j in range(n_qubits)]

    phase = None
    if tunable_phase:
        phase = qd.VariationalParameter("phase")

    return chain(
        rydberg_hea_layer(
            register,
            VariationalParameter(f"At{prefix}_{layer}"),
            VariationalParameter(f"Omega{prefix}_{layer}"),
            VariationalParameter(f"wait{prefix}_{layer}"),
            detunings=detunings,
            drives=drives,
            phase=phase,
        )
        for layer in range(n_layers)
    )



from __future__ import annotations

import re
from typing import Callable

import torch
from qadence import QNN, Parameter, QuantumCircuit, QuantumModel
from qadence.logger import get_logger
from torch.optim.optimizer import Optimizer, required

from qadence_model.optimizers.qfi import (
    get_quantum_fisher,
    get_quantum_fisher_spsa,
)
from qadence_model.types import FisherApproximation

logger = get_logger(__name__)


def _identify_circuit_vparams(
    model: QuantumModel | QNN, circuit: QuantumCircuit
) -> dict[str, Parameter]:
    """Returns the parameters of the model that are circuit parameters.

     Args:
        model (QuantumModel|QNN): The model
        circuit (QuantumCircuit): The quantum circuit

    Returns:
        dict[str, Parameter]:
            Dictionary containing the circuit parameters
    """
    non_circuit_vparams = []
    circ_vparams = {}
    pattern = r"_params\."
    for n, p in model.named_parameters():
        n = re.sub(pattern, "", n)
        if p.requires_grad:
            if n in circuit.parameters():
                circ_vparams[n] = p
            else:
                non_circuit_vparams.append(n)

    if len(non_circuit_vparams) > 0:
        msg = f"""Parameters {non_circuit_vparams} are trainable parameters of the model
                 which are not part of the quantum circuit. Since the QNG optimizer can
                 only optimize circuit parameters, these parameter will not be optimized.
                 Please use another optimizer for the non-circuit parameters."""
        logger.warning(msg)

    return circ_vparams


class QuantumNaturalGradient(Optimizer):
    """Implements the Quantum Natural Gradient Algorithm.

    There are currently two variants of the algorithm implemented: exact QNG and
    the SPSA approximation.

    Unlike other torch optimizers, QuantumNaturalGradient does not take a `Sequence`
    of parameters as an argument, but rather the QuantumModel whose parameters are to be
    optimized. All circuit parameters in the QuantumModel will be optimized.

    WARNING: The exact QNG optimizer is very inefficient both in time and memory as
    it calculates the exact Quantum Fisher Information of the circuit at every
    iteration. Therefore, it is not meant to be run with medium to large circuits.
    Other approximations such as the SPSA are much more efficient while retaining
    good performance.
    """

    def __init__(
        self,
        model: QuantumModel | QNN = required,
        lr: float = required,
        approximation: FisherApproximation | str = FisherApproximation.SPSA,
        beta: float = 10e-3,
        epsilon: float = 10e-2,
    ):
        """
        Args:

            model (QuantumModel):
                Model whose (circuit) parameters are to be optimized
            lr (float): Learning rate.
            approximation (FisherApproximation):
                Approximation used to compute the QFI matrix. Defaults to FisherApproximation.SPSA
            beta (float):
                Shift applied to the QFI matrix before inversion to ensure numerical stability.
                Defaults to 10e-3.
            epsilon (float):
                Finite difference used when computing the SPSA derivatives. Defaults to 10e-2.
        """

        if 0.0 > lr:
            raise ValueError(f"Invalid learning rate: {lr}")
        if 0.0 >= beta:
            raise ValueError(f"Invalid beta value: {beta}")
        if 0.0 > epsilon:
            raise ValueError(f"Invalid epsilon value: {epsilon}")

        if not isinstance(model, QuantumModel):
            raise TypeError(
                f"The model should be an instance of '<class QuantumModel>'. Got {type(model)}."
            )

        self.model = model
        self.circuit = model._circuit.abstract
        if not isinstance(self.circuit, QuantumCircuit):
            raise TypeError(
                f"""The circuit should be an instance of '<class QuantumCircuit>'.
                Got {type(self.circuit)}"""
            )

        circ_vparams = _identify_circuit_vparams(model, self.circuit)
        self.vparams_keys = list(circ_vparams.keys())
        vparams_values = list(circ_vparams.values())

        defaults = dict(
            lr=lr,
            approximation=approximation,
            beta=beta,
            epsilon=epsilon,
        )

        super().__init__(vparams_values, defaults)

        if len(self.param_groups) != 1:
            raise ValueError("QNG doesn't support per-parameter options (parameter groups)")

        if approximation == FisherApproximation.SPSA:
            state = self.state
            state.setdefault("iter", 0)
            state.setdefault("qfi_estimator", None)

    def step(self, closure: Callable | None = None) -> torch.Tensor:
        """Performs a single optimization step of the QNG algorithm.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        assert len(self.param_groups) == 1
        group = self.param_groups[0]

        approximation = group["approximation"]
        beta = group["beta"]
        epsilon = group["epsilon"]
        lr = group["lr"]
        circuit = self.circuit
        vparams_keys = self.vparams_keys
        vparams_values = group["params"]
        grad_vec = torch.tensor([v.grad.data for v in vparams_values])

        if approximation == FisherApproximation.EXACT:
            qng_exact(vparams_values, vparams_keys, grad_vec, lr, circuit, beta)
        elif approximation == FisherApproximation.SPSA:
            qng_spsa(
                vparams_values,
                vparams_keys,
                grad_vec,
                lr,
                circuit,
                self.state,
                epsilon,
                beta,
            )
        else:
            raise NotImplementedError(
                f"""Approximation {approximation} of the QNG optimizer
                is not implemented. Choose an item from the
                FisherApproximation enum: {FisherApproximation.list()}."""
            )

        return loss


def qng_exact(
    vparams_values: list,
    vparams_keys: list,
    grad_vec: torch.Tensor,
    lr: float,
    circuit: QuantumCircuit,
    beta: float,
) -> None:
    """Functional API that performs exact QNG algorithm computation.

    See :class:`~qadence_gradient.qinfo_tools.QuantumNaturalGradient` for details.
    """

    # EXACT metric tensor
    vparams_dict = dict(zip(vparams_keys, vparams_values))
    metric_tensor = 0.25 * get_quantum_fisher(
        circuit,
        vparams_dict=vparams_dict,
    )
    with torch.no_grad():
        # Apply a finite shift to the metric tensor to avoid numerical
        # stability issues when solving the least squares problem
        metric_tensor = metric_tensor + beta * torch.eye(len(grad_vec))

        # Get transformed gradient vector solving the least squares problem
        transf_grad = torch.linalg.lstsq(
            metric_tensor,
            grad_vec,
            driver="gelsd",
        ).solution

        for i, p in enumerate(vparams_values):
            p.data.add_(transf_grad[i], alpha=-lr)


def qng_spsa(
    vparams_values: list,
    vparams_keys: list,
    grad_vec: torch.Tensor,
    lr: float,
    circuit: QuantumCircuit,
    state: dict,
    epsilon: float,
    beta: float,
) -> None:
    """Functional API that performs the QNG-SPSA algorithm computation.

    See :class:`~qadence_gradient.qinfo_tools.QuantumNaturalGradient` for details.
    """
    with torch.no_grad():
        # Get estimation of the QFI matrix
        vparams_dict = dict(zip(vparams_keys, vparams_values))
        qfi_estimator, qfi_mat_positive_sd = get_quantum_fisher_spsa(
            circuit=circuit,
            iteration=state["iter"],
            vparams_dict=vparams_dict,
            previous_qfi_estimator=state["qfi_estimator"],
            epsilon=epsilon,
            beta=beta,
        )

        # Get transformed gradient vector solving the least squares problem
        transf_grad = torch.linalg.lstsq(
            0.25 * qfi_mat_positive_sd,
            grad_vec,
            driver="gelsd",
        ).solution

        for i, p in enumerate(vparams_values):
            if p.grad is None:
                continue
            p.data.add_(transf_grad[i], alpha=-lr)

        state["iter"] += 1
        state["qfi_estimator"] = qfi_estimator



from __future__ import annotations

from qadence.types import StrEnum

__all__ = ["BasisSet", "ReuploadScaling", "FisherApproximation"]


class BasisSet(StrEnum):
    """Basis set for feature maps."""

    FOURIER = "Fourier"
    """Fourier basis set."""
    CHEBYSHEV = "Chebyshev"
    """Chebyshev polynomials of the first kind."""


class ReuploadScaling(StrEnum):
    """Scaling for data reuploads in feature maps."""

    CONSTANT = "Constant"
    """Constant scaling."""
    TOWER = "Tower"
    """Linearly increasing scaling."""
    EXP = "Exponential"
    """Exponentially increasing scaling."""


class FisherApproximation(StrEnum):
    """Approximation to calculate the Quantum Fisher Information (QFI)."""

    EXACT = "Exact"
    """No approximation, computes the exact QFI."""
    SPSA = "SPSA"
    """Approximation via the SPSA algorithm."""



# flake8: noqa

from .feature_maps import (
    feature_map,
    exp_fourier_feature_map,
)

from .hea import hea

from .iia import identity_initialized_ansatz

from .rydberg_hea import rydberg_hea, rydberg_hea_layer
from .rydberg_feature_maps import (
    rydberg_feature_map,
    analog_feature_map,
    rydberg_tower_feature_map,
)

from .qft import qft

from .qnn_config import FeatureMapConfig, AnsatzConfig
from .qnn_constructors import (
    create_fm_blocks,
    create_ansatz,
    create_observable,
    build_qnn_from_configs,
)

from .qnn_model import QNN
from .qcnn_model import QCNN

# Modules to be automatically added to the qadence namespace
__all__ = [
    "feature_map",
    "exp_fourier_feature_map",
    "hea",
    "identity_initialized_ansatz",
    "qft",
    "rydberg_hea",
    "rydberg_hea_layer",
    "rydberg_feature_map",
    "analog_feature_map",
    "rydberg_tower_feature_map",
    "FeatureMapConfig",
    "AnsatzConfig",
    "create_fm_blocks",
    "create_ansatz",
    "create_observable",
    "build_qnn_from_configs",
    "QNN",
    "QCNN",
]



from __future__ import annotations

from importlib import import_module

"""Fetch the functions defined in the __all__ of each sub-module.

Import to the qadence name space. Make sure each added submodule has the respective definition:

    - `__all__ = ["function0", "function1", ...]`

Furthermore, add the submodule to the list below to automatically build
the __all__ of the qadence namespace. Make sure to keep alphabetical ordering.
"""

list_of_submodules = [
    ".constructors",
    ".optimizers",
]

__all__ = []
for submodule in list_of_submodules:
    __all_submodule__ = getattr(import_module(submodule, package="qadence_model"), "__all__")
    __all__ += __all_submodule__
